import numpy
import random
import multiprocessing
from typing import Union
from Supplychain.Wrappers.simulator import CosmoEngine
from Supplychain.Generic.adx_and_file_writer import ADXAndFileWriter
import pandas as pd
from Supplychain.Generic.timer import Timer
from math import sqrt
from Supplychain.Wrappers.environment_variables import EnvironmentVariables
from Supplychain.Run.simulation import run_simple_simulation


def run_simulation_with_seed(simulation_name: str,
                             seed: int,
                             amqp_consumer_adress: Union[str, None] = None,
                             output_list=None) -> bool:
    # Reduce log level to Error during optimization
    logger = CosmoEngine.LoggerManager.GetInstance().GetLogger()
    logger.SetLogLevel(logger.eAssertion)
    simulator = CosmoEngine.LoadSimulator('Simulation')

    simulator.FindAttribute("{Model}Model::{Attribute}Seed").SetAsString(
        str(seed)
    )

    used_probes = ['PerformanceIndicators', 'Stocks']
    used_consumers = ['CSVPerformanceIndicatorsConsumer', 'CSVStocksConsumer']

    # Delete all probes expect for the one we use
    for probe in simulator.GetProbes():
        if probe.GetType() not in used_probes:
            simulator.DestroyProbe(probe)

    if amqp_consumer_adress is not None:
        # Instantiate consumers using AMQP to send data to the cloud service
        simulator.InstantiateAMQPConsumers(simulation_name, amqp_consumer_adress)

        used_consumers = ['PerformanceIndicatorsAMQP']

    # Remove unused consumers
    for consumer in simulator.GetConsumers():
        if consumer.GetName() not in used_consumers:
            simulator.DestroyConsumer(consumer)

    class StockConsumer(CosmoEngine.Consumer):
        memory = list()

        def Consume(self, p_data):
            probe_output = CosmoEngine.StocksProbeOutput.Cast(p_data)
            f = probe_output.GetFacts()
            timestep = int(probe_output.GetProbeRunDimension().GetProbeOutputCounter())
            for data in f:
                fact = [str(data.GetAttributeAsString('ID')),
                        timestep,
                        float(data.GetAttributeAsFloat64('Demand')),
                        float(data.GetAttributeAsFloat64('RemainingQuantity')),
                        float(data.GetAttributeAsFloat64('ServedQuantity')),
                        float(data.GetAttributeAsFloat64('UnservedQuantity'))]
                self.memory.append(fact)

    consumer = StockConsumer("LocalConsumer")
    consumer.Connect(simulator.GetProbe("Stocks"))

    # Run simulation
    simulator.Run()

    if output_list is not None:
        output_list.extend(StockConsumer.memory)

    # Remove all the consumers in case that amqp consumers are still connected to ADX
    for consumer in simulator.GetConsumers():
        simulator.DestroyConsumer(consumer)

    return simulator.IsFinished()


def uncertainty_analysis(simulation_name: str,
                         amqp_consumer_adress: Union[str, None] = None,
                         sample_size: int = 1000,
                         batch_size: int = 100,
                         adx_writer: Union[ADXAndFileWriter, None] = None):
    with Timer('[Run Uncertainty]') as t:
        if batch_size > sample_size:
            batch_size = sample_size

        maxint = numpy.iinfo(numpy.int32).max
        seedlist = random.sample(range(maxint), sample_size)
        processes_size = min(multiprocessing.cpu_count(), batch_size)
        manager = multiprocessing.Manager()
        probe_data = manager.list()
        t.display_message("Starting simulations")
        with multiprocessing.Pool(processes_size) as p:
            for i in range(0, len(seedlist), batch_size):
                subseedlist = seedlist[i:i + batch_size]
                params = list(map(lambda seed: (simulation_name, seed, amqp_consumer_adress, probe_data), subseedlist))
                p.starmap(run_simulation_with_seed, params)
        t.split("Ended simulations : {time_since_start}")

        df = pd.DataFrame((l for l in probe_data))
        df.columns = ['StockId', 'TimeStep', 'Demand', 'RemainingQuantity', 'ServedQuantity', 'UnservedQuantity']
        t.split("Create dataframes for stats computation: {time_since_last_split}")

        quantile_list = [0.05, 0.25, 0.5, 0.75, 0.95]
        groups = df.groupby(['StockId', 'TimeStep'])
        df_quantiles = groups.quantile(quantile_list)
        df_average = groups.mean()
        df_error = groups.agg(lambda x: x.std() / sqrt(x.count()))
        t.split("Compute stats : {time_since_last_split}")

        # since use of pivot to have 1 column per tuple (quantile, valuetype)
        # then use of stack to have 1 line per (stock, timestep, valuetype)
        df3 = df_quantiles.reset_index().pivot(index=['StockId',
                                                      'TimeStep'],
                                               columns='level_2',
                                               values=['Demand',
                                                       'RemainingQuantity',
                                                       'ServedQuantity',
                                                       'UnservedQuantity']).stack(level=0)
        df3.reset_index(inplace=True)
        df3.columns = ['StockId',
                       'TimeStep',
                       'Category'] + [f'Percentile{int(v * 100)}'
                                      for v in quantile_list]

        # use of stack to have 1 line per (stock, timestep, valuetype)
        df4 = df_average.stack(level=0)
        df4 = df4.reset_index()
        df4.columns = ['StockId',
                       'TimeStep',
                       'Category',
                       'Mean']

        # use of stack to have 1 line per (stock, timestep, valuetype)
        df5 = df_error.stack(level=0)
        df5 = df5.reset_index()
        df5.columns = ['StockId',
                       'TimeStep',
                       'Category',
                       'SE']

        # Merge of dfs to final df
        final_df = pd.merge(df3, df4, on=['StockId',
                                          'TimeStep',
                                          'Category'])
        final_df = pd.merge(final_df, df5, on=['StockId',
                                               'TimeStep',
                                               'Category'])

        final_df['SimulationRun'] = EnvironmentVariables.simulation_id
        final_df = final_df[['TimeStep',
                             'SimulationRun',
                             'StockId',
                             'Percentile5',
                             'Percentile25',
                             'Percentile50',
                             'Percentile75',
                             'Percentile95',
                             'Mean',
                             'SE',
                             'Category']]
        adx_writer.write_target_file(final_df.to_dict('records'), 'StockUncertaintiesStatistics', EnvironmentVariables.simulation_id)

        t.split("Sent stats to ADX : {time_since_last_split}")
        t.display_message("Running simple simulation to fill ADX")
        # Put back log level to Info for final simulation
        # Reduce log level to Error during optimization
        logger = CosmoEngine.LoggerManager.GetInstance().GetLogger()
        logger.SetLogLevel(logger.eInfo)

        stop_uncertainty = {
            "Model::@ActivateUncertainties": "false"
        }

        run_simple_simulation(simulation_name=simulation_name,
                              amqp_consumer_adress=amqp_consumer_adress,
                              modifications=stop_uncertainty)



simulator_format = {
    "ModelParameters": [
        "TimeStepDuration",
        "TimeStepPerCycle",
        "NumberOfCycle",
        "StartingDate",
        "ManageBacklogQuantities",
        "OptimizationObjective",
        "ActivateUncertainties",
        "EmptyObsoleteStocks",
        "UncertaintiesProbabilityDistribution",
        "TransportUncertaintiesProbabilityDistribution",
        "ActivateVariableMachineOpeningRate",
        "ImmobilizedCashRelativeCost",
        "BatchSize",
        "ExtraProductionPlanIsAvailable",
        "EnforceProductionPlan",
        "ActualizeShipments",
        "ActivateCorrelatedDemandUncertainties",
        "DemandCorrelations",
    ],
    "BE_Stock": [
        "Name",
        "CurrentStock",
        "MinimalStock",
        "MaximalStock",
        "Value",
        "PurchasingUnitCostSchedule",
        "CO2UnitEmissionsSchedule",
        "UnitIncomeSchedule",
        "Demand",
        "StorageUnitCostSchedule",
        "IsInfinite",
        "StockPolicy",
        "SourcingPolicy",
        "DispatchPolicy",
        "ReviewPeriod",
        "FirstReview",
        "Advance",
        "LeadTime",
        "SalesForecasts",
        "OrderPointSchedule",
        "OrderQuantitySchedule",
        "SafetyQuantitySchedule",
        "SchedulingStep",
    ],
    "CE_Machine": [
        "Name",
        "StockTypeProportion",
        "Groups",
        "OpeningRates",
        "OpeningTimeSchedule",
        "ProductionCostSchedule",
        "ProductionPolicy",
    ],
    "BE_Operation": [
        "Name",
        "ParentName",
        "OperatingPerformanceSchedule",
        "CycleTimeSchedule",
        "RejectRateSchedule",
        "OperatingPerformanceUncertaintySchedule",
        "ProductionUnitCostSchedule",
        "CO2UnitEmissionsSchedule",
        "ProductionPlanSchedule",
        "InvestmentCost",
        "IsContractor",
        "Priority",
        "SchedulingStep",
        "FollowsStockPolicy",
        "Duration",
        "LeadTime",
        "MinimumOrderQuantitySchedule",
        "MultipleOrderQuantitySchedule",
        "SourcingProportionSchedule",
    ],
    "BE_TransportOperation": [
        "Name",
        "Duration",
        "Shipments",
        "TransportUnitCostSchedule",
        "CustomsUnitCostSchedule",
        "CO2UnitEmissionsSchedule",
        "Priority",
        "SchedulingStep",
        "FollowsStockPolicy",
        "LeadTime",
        "ActualDurationSchedule",
        "MinimumOrderQuantitySchedule",
        "MultipleOrderQuantitySchedule",
        "SourcingProportionSchedule",
        "TransportUncertaintiesParameter1",
        "TransportUncertaintiesParameter2",
        "TransportUncertaintiesParameter3",
        "TransportUncertaintiesParameter4",
    ],
    "EE_StockToOperation": [
        "Name",
        "OperationName",
        "StockName",
        "PropStockForOperation",
        "IsProportion",
        "Groups",
        "Cardinality",
    ],
    "Arcs_OperationToStock": [
        "StockName",
        "OperationName",
    ],
    "IndustrialNetworkScheduler": [
        "Name",
        "Type",
        "TaskName",
    ],
}



import csv
from Supplychain.Generic.folder_io import FolderWriter

from typing import Union


class CSVWriter(FolderWriter):

    def write_from_list(self, dict_list: list, file_name: str, ordering_key: Union[str, None] = None):
        with open(f"{self.output_folder}/{file_name}.csv", "w") as target_file:
            if dict_list:
                fieldnames = list(dict_list[0].keys())
                fieldnames.extend(
                    set(fieldname for row in dict_list for fieldname in row)
                    - set(fieldnames)
                )
                writer = csv.DictWriter(target_file, fieldnames=fieldnames, restval=None)
                writer.writeheader()
                to_be_writen = dict_list
                if ordering_key is not None and ordering_key in dict_list[0]:
                    to_be_writen = sorted(dict_list, key=lambda e: e[ordering_key])
                for row in to_be_writen:
                    to_be = {k: FolderWriter.json_value(v)
                             for k, v
                             in row.items()}
                    writer.writerow(to_be)

    def __init__(self,
                 output_folder: str = "Output"):
        FolderWriter.__init__(self,
                              output_folder=output_folder)



from azure.storage.queue import QueueClient, TextBase64EncodePolicy
import json
from Supplychain.Generic.timer import Timer


class QueueWriter(Timer):

    def send_items(self, items):
        total_messages = len(items)
        for item in items:
            message = json.dumps(item, separators=(',', ':'))
            self.queue_client.send_message(message)
        self.display_message(f"Sent {total_messages} message{'s' if total_messages > 1 else ''} to the queue.")

    def __init__(self, connect_string: str, queue_name: str):
        Timer.__init__(self, "[Storage Queue Writer]")
        self.queue_client = QueueClient.from_connection_string(connect_string,
                                                               queue_name=queue_name,
                                                               message_encode_policy=TextBase64EncodePolicy())



from Supplychain.Protocol.protocol import AbstractObjectiveFunction


class ProfitMaximizationObjectiveFunction(AbstractObjectiveFunction):
    """
    Profit Maximization objective function :
    We compute the maximal profit for the model.
    """
    def applies(self, simulation_result):

        objective = -sum(map(lambda r: float(r.get('Profit', 0)), simulation_result))

        return objective


class DefaultObjectiveFunction(AbstractObjectiveFunction):
    """
    Default objective function for SupplyChain :
    We compute a distance to demand valuation of the results
    """
    def applies(self, simulation_result):

        objective = sum(map(lambda r: float(r.get('ServiceLevelIndicator', 0)), simulation_result))

        return objective



import multiprocessing
from typing import Union
from copy import deepcopy
import pandas
from Supplychain.Wrappers.simulator import CosmoEngine
from Supplychain.Generic.adx_and_file_writer import ADXAndFileWriter
from Supplychain.Generic.timer import Timer
from Supplychain.Wrappers.environment_variables import EnvironmentVariables
from Supplychain.Run.simulation import run_simple_simulation
import comets as co


def uncertainty_analysis(
    simulation_name: str,
    simulation_path: str = "Simulation",
    amqp_consumer_adress: Union[str, None] = None,
    sample_size: int = 1000,
    batch_size: int = 100,
    adx_writer: Union[ADXAndFileWriter, None] = None,
    validation_folder: Union[str, None] = None,
):

    with Timer("[Run Uncertainty]") as t:
        if batch_size > sample_size:
            batch_size = sample_size

        processes_size = min(multiprocessing.cpu_count(), batch_size)

        used_probes = ["PerformanceIndicators", "Stocks"]
        used_consumers = ["PerformanceIndicatorsAMQP"]

        class StockConsumer:
            def __init__(self):
                self.memory = list()

            def Consume(self, p_data):
                probe_output = self.engine.StocksProbeOutput.Cast(p_data)
                f = probe_output.GetFacts()
                timestep = int(
                    probe_output.GetProbeRunDimension().GetProbeOutputCounter()
                )
                for data in f:
                    fact = [
                        str(data.GetAttributeAsString("ID")),
                        timestep,
                        float(data.GetAttributeAsFloat64("Demand")),
                        float(data.GetAttributeAsFloat64("RemainingQuantity")),
                        float(data.GetAttributeAsFloat64("ServedQuantity")),
                        float(data.GetAttributeAsFloat64("UnservedQuantity")),
                    ]
                    self.memory.append(fact)

        class PerformanceConsumer:
            """
            Python Consumer for performance indicators.
            Currently not used by the solution (results are also sent directly to ADX through the PerformanceIndicatorsAMQP consumer),
            but available for the future and for local computation of performance indicators statistics.
            """

            def __init__(self):
                self.memory = list()

            def Consume(self, p_data):
                probe_output = self.engine.PerformanceIndicatorsProbeOutput.Cast(p_data)
                f = probe_output.GetFacts()
                for data in f:
                    fact = {
                        "OPEX": float(data.GetAttributeAsFloat64("OPEX")),
                        "Profit": float(data.GetAttributeAsFloat64("Profit")),
                        "AverageStockValue": float(
                            data.GetAttributeAsFloat64("AverageStockValue")
                        ),
                        "ServiceLevelIndicator": float(
                            data.GetAttributeAsFloat64("ServiceLevelIndicator")
                        ),
                        "CO2Emissions": float(
                            data.GetAttributeAsFloat64("CO2Emissions")
                        ),
                        "TotalDemand": float(data.GetAttributeAsFloat64("TotalDemand")),
                        "TotalServedQuantity": float(
                            data.GetAttributeAsFloat64("TotalServedQuantity")
                        ),
                    }
                    self.memory.append(fact)

        size_of_performance_consumer = 7  # Number of KPIs in PerformanceConsumer

        simulator_interface = co.CosmoInterface(
            simulator_path=simulation_path,
            custom_sim_engine=CosmoEngine,
            simulation_name=simulation_name,
            amqp_consumer_address=amqp_consumer_adress,
            used_consumers=used_consumers,
            used_probes=used_probes,
            custom_consumers=[
                (StockConsumer, "LocalConsumer", "Stocks"),
                (
                    PerformanceConsumer,
                    "LocalPerformanceConsumer",
                    "PerformanceIndicators",
                ),
            ],
        )

        # Load simulator to be able to access attributes of the model
        simulator_interface.initialize()

        # Getting the name of all the transport operations
        all_transports = get_transports(simulator_interface.sim)

        # Retrieving model information
        (
            max_time_step,
            distribution,
            actual_duration_schedule,
            list_of_transports,
            uncertainty_param_1,
            uncertainty_param_2,
            uncertainty_param_3,
            uncertainty_param_4,
            demands,
            ActivateCorrelatedDemandUncertainties,
            DemandCorrelations,
            uncertain_stocks,
        ) = model_info(simulator_interface, simulation_path, all_transports)

        simulator_interface.terminate()

        distribution_params = get_distribution_parameter(distribution)

        # Extending the dictionaries above (scheduled attributes)
        extended_actual_duration = extend_dic(actual_duration_schedule, max_time_step)
        extended_param_1 = extend_dic(uncertainty_param_1, max_time_step)
        extended_param_2 = extend_dic(uncertainty_param_2, max_time_step)
        extended_param_3 = extend_dic(uncertainty_param_3, max_time_step)
        extended_param_4 = extend_dic(uncertainty_param_4, max_time_step)
        extended_demands = extend_dic(demands, max_time_step)

        def encoder(parameters):
            """
            The encoder takes a parameterset containing the input parameters that are changing at each simulation.
            It has the following format :
            {'{Model} [..] Seed': 3251851028, 'T_@_0': 1.7504164949122698, 'T_@_1': 3.05256524351985, [...], 'T_@_10': 3.2371899035525793, 'StockA':[1.1, 2., ..., 3.4]}
            The parameters correspond either to:
            - the Seed datapath and its value
            - transport_name +_@_ + time_step, and the duration of this transport at this time step
            - stock_name, and a list of demands for this stock

            It returns a parameterset where the keys are datapaths in the model.
            The values for each transport's duration is transformed in one dictionary {TimeStep: value} and updated as follows:
            new_transport_duration = max(0, round(old_transport_duration + sample input drawn from the distribution))
            The values for each stock's demand is set to its ExternalDemand in the attribute Demand, which is a dictionary {TimeStep: Composite attribute}.
            {'Model [..] T::@ActualDurationSchedule': {'0': 5, '1': 6, [..] '10': 44}, '{Model}Model::{Attribute}Seed': 3209521878, ...}
            """
            encoded_parameterset = {
                "{Model}Model::{Attribute}Seed": parameters[
                    "{Model}Model::{Attribute}Seed"
                ],
            }
            for transports in list_of_transports:
                ActualDurationSchedule = {}
                for i in range(len(extended_actual_duration[transports])):
                    ActualDurationSchedule[str(i)] = max(
                        0,
                        round(
                            extended_actual_duration[transports][i]
                            + parameters[f"{transports}_@_{i}"]
                        ),
                    )
                encoded_parameterset[
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transports}::@ActualDurationSchedule"
                ] = ActualDurationSchedule
            for stock in uncertain_stocks:
                sample_demand = parameters[f"{stock}"]
                demand_attribute = deepcopy(extended_demands[f"{stock}"])
                for i in range(max_time_step):
                    demand_attribute[i]["ExternalDemand"] = sample_demand[i]
                encoded_parameterset[
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{stock}::@Demand"
                ] = demand_attribute
            return encoded_parameterset

        def get_outcomes(modelinterface):
            """
            Returns a parameter set with all the model's output. More precisely, the parameter set is the
            the result of the concatenation of three parameter sets.

            The first one looks like this:
            {'U__&@&__0': 5, 'U__&@&__1': 5, [...], 'U__&@&__10': 6}. The keys correspond to transport_name + __&@&__ + time_step,
            and the value to the duration of the transport at this time step.

            The second parameter set looks like this:
            {'A__&@&__ServedQuantity__&@&__0': 0.0, 'A__&@&__UnservedQuantity__&@&__0': 0.0}. The keys correspond to stock + __&@&__ + category + time_step

            The third parameter set contains the performance indicators (OPEX, Profit, ...).
            """
            transport_duration = get_transport_duration(
                all_transports, modelinterface, max_time_step
            )
            transports_and_consumer = {
                **transport_duration,
                **transform_consumer_memory(modelinterface.LocalConsumer.memory),
            }
            performances = modelinterface.LocalPerformanceConsumer.memory[0]
            transports_and_consumer.update(performances)
            return transports_and_consumer

        if (
            ActivateCorrelatedDemandUncertainties
        ):  # Correlated demands are not compatible with demands drawn inside the model
            cold_input_parameter_set = {
                "{Model}Model::{Attribute}ActivateUncertainties": 0
            }
        else:
            cold_input_parameter_set = {}

        simulationtask = co.ModelTask(
            modelinterface=simulator_interface,
            encode=encoder,
            get_outcomes=get_outcomes,
            cold_input_parameter_set=cold_input_parameter_set,
        )

        # Creating the uncertainty analysis sampling
        sampling = creating_transport_distribution_space(
            distribution,
            extended_param_1,
            extended_param_2,
            extended_param_3,
            extended_param_4,
            list_of_transports,
            distribution_params,
            max_time_step,
        )
        # adding the seed as an uncertain parameter
        sampling.append(
            {
                "name": "{Model}Model::{Attribute}Seed",
                "sampling": "seed_generator",
            }
        )

        # adding the generator of samples on the demand
        if ActivateCorrelatedDemandUncertainties:
            sampling += creating_demand_generator(
                extended_demands, max_time_step, DemandCorrelations
            )

        if validation_folder is not None:
            save_task_history = True
        else:
            save_task_history = False

        experiment = co.UncertaintyAnalysis(
            task=simulationtask,
            sampling=sampling,
            stop_criteria={"max_evaluations": sample_size},
            analyzer=["standard", "quantiles"],
            n_jobs=processes_size,
            save_task_history=save_task_history,
        )

        t.display_message("Starting simulations")
        experiment.run()
        t.split("Ended simulations : {time_since_start}")

        # Separating the results data on the stock from the results data on transport's duration from the results data on performances
        transport_limiter = len(all_transports) * max_time_step
        df_transport_duration = experiment.results["statistics"].iloc[
            0:transport_limiter, :
        ]  # First transport_limiter rows are results about transport durations
        df_probe_data = experiment.results["statistics"].iloc[
            transport_limiter:-size_of_performance_consumer, :
        ]  # Next are results about stocks levels
        # Last rows are general performances, not sent to ADX since the performance indicators are also sent directly by the AMQP consumer
        # This object is unused for now, but might be used in the future
        performances = experiment.results["statistics"].iloc[
            -size_of_performance_consumer:, :
        ]  # noqa

        df_probe_data.reset_index(inplace=True)
        df_transport_duration.reset_index(inplace=True)

        # Transforming the data on both the stock and the transport duration
        df_stock_final = transform_stock_data(df_probe_data)
        df_transport_final = transform_transport_data(df_transport_duration)

        if validation_folder:
            # Writing the results locally in csv files

            # Get all demands directly from the experiment, before aggregation of statistics
            demands = []
            j = 0
            for i in experiment.task_history["outputs"]:

                for (k, v) in i.items():
                    if "__&@&__Demand__&@&__" in k:
                        demand_result_dict = {}
                        demand_result_dict["Simulation"] = j
                        demand_result_dict["Entity"] = k.split("__&@&__Demand__&@&__")[
                            0
                        ]
                        demand_result_dict["TimeStep"] = k.split(
                            "__&@&__Demand__&@&__"
                        )[1]
                        demand_result_dict["Demand"] = v

                        demands.append(demand_result_dict)
                j += 1
            demand_df = pandas.DataFrame(demands)
            demand_df.to_csv(
                str(validation_folder) + "/df_all_demands.csv", index=False
            )
            df_stock_final.to_csv(str(validation_folder) + "/final_df_comets.csv")
            df_transport_final.to_csv(str(validation_folder) + "/df_transport.csv")
            performances.to_csv(str(validation_folder) + "/df_performances.csv")

        if adx_writer is not None:
            adx_writer.write_target_file(
                df_stock_final.to_dict("records"), "StockUncertaintiesStatistics"
            )
            adx_writer.write_target_file(
                df_transport_final.to_dict("records"), "TransportUncertaintyStatistics"
            )
            t.split("Sent stats to ADX : {time_since_last_split}")
        t.display_message("Running simple simulation to fill ADX")
        # Put back log level to Info for final simulation
        # Reduce log level to Error during optimization
        logger = CosmoEngine.LoggerManager.GetInstance().GetLogger()
        logger.SetLogLevel(logger.eInfo)

        stop_uncertainty = {"Model::@ActivateUncertainties": "false"}

        run_simple_simulation(
            simulation_name=simulation_name,
            simulation_path=simulation_path,
            amqp_consumer_adress=amqp_consumer_adress,
            modifications=stop_uncertainty,
        )


# Function to get the list of all the transports in the simulation
def get_transports(temporary_simulator):
    transports_list = []
    transports = temporary_simulator.get_entities_names_by_type(
        entity_type="TransportOperation"
    )
    for keys in transports:
        transports_list.append(keys)
    return transports_list


# Function to get the list of all the stocks that have uncertain demand
def get_stocks(temporary_simulator):
    uncertain_stocks = []
    for stock in temporary_simulator.get_entities_by_type("Stock"):
        demands = CosmoEngine.DataTypeMapInterface.Cast(stock.GetAttribute("Demand"))
        stock_name = stock.GetName()
        for time_step in demands.GetKeys():
            demand = demands.GetAt(time_step)
            if demand.GetAttribute("DemandRelativeUncertainty").Get() > 0:
                uncertain_stocks.append(stock_name)
                break
    return uncertain_stocks


def get_distribution_parameter(my_distribution):
    return co.DistributionRegistry.information[str(my_distribution)]["parameters"]


# Function to get the model informations for each transports, such as the distribution used and
# its parameters, and for each stock, such as which stocks have uncertain demand
def model_info(my_simulator, simulation_path, transports_names):
    actual_duration_schedule = {}
    list_of_transports = transports_names.copy()
    uncertainty_param_1 = {}
    uncertainty_param_2 = {}
    uncertainty_param_3 = {}
    uncertainty_param_4 = {}
    demands = {}

    time_step_per_cycle = my_simulator.get_outputs(["Model::@TimeStepPerCycle"])[
        "Model::@TimeStepPerCycle"
    ]

    number_of_cycle = my_simulator.get_outputs(["Model::@NumberOfCycle"])[
        "Model::@NumberOfCycle"
    ]

    max_time_step = time_step_per_cycle * number_of_cycle

    distribution = my_simulator.get_outputs(
        ["Model::@TransportUncertaintiesProbabilityDistribution"]
    )["Model::@TransportUncertaintiesProbabilityDistribution"]

    ActivateCorrelatedDemandUncertainties = my_simulator.get_outputs(
        ["Model::@ActivateCorrelatedDemandUncertainties"]
    )["Model::@ActivateCorrelatedDemandUncertainties"]

    if ActivateCorrelatedDemandUncertainties:
        uncertain_stocks = get_stocks(my_simulator.sim)
    else:
        uncertain_stocks = []

    DemandCorrelations = my_simulator.get_outputs(["Model::@DemandCorrelations"])[
        "Model::@DemandCorrelations"
    ]

    for transport in transports_names:

        # If their is no ActualDuration in the TransportSchedules column of the dataset,
        # we use by default the attribute duration, in the Transport column of the dataset.
        if (
            my_simulator.get_outputs(
                [
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
                ]
            )[
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
            ]
            == {}
        ):
            duration = my_simulator.get_outputs(
                [
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@Duration"
                ]
            )[f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@Duration"]

            # Use a dict format so that the function "extended_dict" defined below can be applied
            actual_duration_schedule[transport] = {0: duration}
        else:
            actual_duration_schedule[transport] = my_simulator.get_outputs(
                [
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
                ]
            )[
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
            ]

        uncertainty_param_1[transport] = my_simulator.get_outputs(
            [
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter1"
            ]
        )[
            f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter1"
        ]
        uncertainty_param_2[transport] = my_simulator.get_outputs(
            [
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter2"
            ]
        )[
            f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter2"
        ]
        uncertainty_param_3[transport] = my_simulator.get_outputs(
            [
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter3"
            ]
        )[
            f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter3"
        ]
        uncertainty_param_4[transport] = my_simulator.get_outputs(
            [
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter4"
            ]
        )[
            f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@TransportUncertaintiesParameter4"
        ]
        # If the transport has no parameters, its transport duration will not be part of the uncertainty analysis
        if (
            uncertainty_param_1[transport] == uncertainty_param_2[transport]
            and uncertainty_param_1[transport] == {}
        ):
            list_of_transports.remove(transport)

    for stock in uncertain_stocks:
        demands[stock] = my_simulator.get_outputs(
            [f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{stock}::@Demand"]
        )[f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{stock}::@Demand"]

    return (
        max_time_step,
        distribution,
        actual_duration_schedule,
        list_of_transports,
        uncertainty_param_1,
        uncertainty_param_2,
        uncertainty_param_3,
        uncertainty_param_4,
        demands,
        ActivateCorrelatedDemandUncertainties,
        DemandCorrelations,
        uncertain_stocks,
    )


# Function to extend dictionaries using transport schedules.
# the following dic: {0: 3, 6: 4, 7: 3, 8: 2, 9: 8, 10: 40}
# Becomes: {0: 3, 1: 3, 2: 3, 3: 3, 4: 3, 5: 3, 6: 4, 7: 3, 8: 2, 9: 8, 10: 40}
def extend_simple_dic(my_dic, number_of_iterations):
    if my_dic != {}:  # checking that the dic isn't empty
        extended_dic = {
            0: my_dic[0]
        }  # We assume that the uncertainty starts at the first time step
        for i in range(1, number_of_iterations):
            if i in my_dic:
                extended_dic[i] = deepcopy(my_dic[i])
            else:
                extended_dic[i] = deepcopy(extended_dic[i - 1])
    else:
        extended_dic = {}
    return extended_dic


# Function to extend dictionaries using schedules.
# For each transports, the following dic: {"transport name" : {0: 3, 6: 4, 7: 3, 8: 2, 9: 8, 10: 40}}
# Becomes: {"transport name" : {0: 3, 1: 3, 2: 3, 3: 3, 4: 3, 5: 3, 6: 4, 7: 3, 8: 2, 9: 8, 10: 40}}
def extend_dic(my_dic, number_of_iterations):
    extended_dic = {}
    for entity in my_dic.keys():
        extended_dic[entity] = extend_simple_dic(my_dic[entity], number_of_iterations)
    return extended_dic


# this function transforms supply chain's distribution parameters
# that don't match CoMETS format
def check_distrib_parameters(
    my_distribution, param1, param2, param3, param4, transports_names
):
    # In CoMETS the upper bound of the discreteuniform distribution is excluded.
    # However, in supply chain it is included. Therefore, we need to add 1 to the upper bound
    # so it matches CoMETS
    if my_distribution == "discreteuniform":
        for transport in transports_names:
            for keys in param2[transport].keys():
                param2[transport][keys] += 1

    # In CoMETS the two arguments of the uniform distribution are loc and scale
    # and the interval of the distribution is the following:  [loc, loc+scale]
    # However, in supply chain the two parameters of the uniform distribution are
    # [lower, upper]. Therefor, upper needs to be mapped to scale
    if my_distribution == "uniform":
        for transport in transports_names:
            for keys in param2[transport].keys():
                param2[transport][keys] = (
                    param2[transport][keys] - param1[transport][keys]
                )

    if my_distribution == "betabinom":
        for transport in transports_names:
            for keys in param1[transport].keys():
                param1[transport][keys] = round(param1[transport][keys])

    if my_distribution == "binomial":
        for transport in transports_names:
            for keys in param1[transport].keys():
                param1[transport][keys] = round(param1[transport][keys])

    if my_distribution == "hypergeom":
        for transport in transports_names:
            for keys in param1[transport].keys():
                param1[transport][keys] = round(param1[transport][keys])
                param2[transport][keys] = round(param2[transport][keys])
                param3[transport][keys] = round(param3[transport][keys])

    return [param1, param2, param3, param4]


# Function to create the uncertainty analysis space according to CoMETS format
# This function will create one variable for each time step of each transports
# distribution: name of the distribution used for the ua
# param1: values by transport by time step for the first parameter of the distribution
# param2: values by transport by time step for the second parameter of the distribution
# param3: values by transport by time step for the third parameter of the distribution
# param4: values by transport by time step for the fourth parameter of the distribution
# transports: list of all the uncertain transports names
# distribution_parameters: list of the parameters names required by CoMETS sampler for the given distribution
# number_of_time_steps: number of time step simulated
def creating_transport_distribution_space(
    distribution,
    param1,
    param2,
    param3,
    param4,
    transports,
    distribution_parameters,
    number_of_time_steps,
):
    sampling = []
    all_parameters = check_distrib_parameters(
        distribution, param1, param2, param3, param4, transports
    )
    for transport in transports:
        considered_parameters = [
            (position, parameter)
            for position, parameter in enumerate(distribution_parameters)
            if all_parameters[position][transport]
        ]
        for t in range(number_of_time_steps):
            parameters = {
                parameter: all_parameters[position][transport][t]
                for position, parameter in considered_parameters
            }
            sampling.append(
                {
                    "name": f"{transport}_@_{t}",
                    "sampling": distribution,
                    "parameters": parameters,
                }
            )
    return sampling


def creating_demand_generator(
    extended_demands, number_of_time_steps, DemandCorrelations
):
    sampling = []
    for stock, demand_attribute in extended_demands.items():
        mean_demand = []
        uncertainties = []
        for t in range(number_of_time_steps):

            demand = demand_attribute[t]["ExternalDemand"]
            mean_demand.append(demand)
            uncertainties.append(
                demand * demand_attribute[t]["DemandRelativeUncertainty"]
            )  # Uncertainty proportional to demand, DemandRelativeUncertainty*Demand is the standard deviation
        sampling.append(
            {
                "name": f"{stock}",
                "sampling": co.TimeSeriesSampler(
                    correlation=DemandCorrelations,
                    dimension=number_of_time_steps,
                    forecast=mean_demand,
                    uncertainties=uncertainties,
                    minimum=0,
                ),
            }
        )
    return sampling


# Function that returns a parameterset with the transport duration for each transport at the end of the simulation
# The transport duration is separated for each time step. The output parameterset (for a simulation
# with 1 TransportOperation: U) will have the following format:
#    {Model[...]U::@ActualDurationSchedule__&@&__0': 10,
#          [...],
#     Model[...]U::@ActualDurationSchedule__&@&__10': 7}
def get_transport_duration(transports_names, modelinterface, max_time_step):
    transport_duration = {}
    transport_duration_transformed = {}
    for transport in transports_names:
        if (
            modelinterface.get_outputs(
                [
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
                ]
            )[
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
            ]
            == {}
        ):
            duration = modelinterface.get_outputs(
                [
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@Duration"
                ]
            )[f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@Duration"]
            # Use a dict format so that the function "extended_dict" defined below can be applied
            actual_duration_schedule = {0: duration}
        else:
            actual_duration_schedule = modelinterface.get_outputs(
                [
                    f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
                ]
            )[
                f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
            ]
        transport_duration[
            f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
        ] = extend_simple_dic(
            actual_duration_schedule,
            max_time_step,
        )
        time_step = 0
        for value in transport_duration[
            f"Model::{{Entity}}IndustrialNetwork::{{Entity}}{transport}::@ActualDurationSchedule"
        ].values():
            transport_duration_transformed[f"{transport}__&@&__{time_step}"] = value
            time_step += 1
    return transport_duration_transformed


# This function transform the consumer memory from a list of list to a list of
# ParameterSet. Note that each sublist in the initial format is transformed into
# a parameterset of 4 parameters : Demand, RemainingQuantity, ServedQuantity,UnservedQuantity
def transform_consumer_memory(transform_consumer_memory):
    dic_of_parameterset = {}
    for elements in transform_consumer_memory:
        dic_of_parameterset[
            str(elements[0] + "__&@&__" + "Demand" + "__&@&__" + str(elements[1]))
        ] = elements[2]
        dic_of_parameterset[
            str(
                elements[0]
                + "__&@&__"
                + "RemainingQuantity"
                + "__&@&__"
                + str(elements[1])
            )
        ] = elements[3]
        dic_of_parameterset[
            str(
                elements[0]
                + "__&@&__"
                + "ServedQuantity"
                + "__&@&__"
                + str(elements[1])
            )
        ] = elements[4]
        dic_of_parameterset[
            str(
                elements[0]
                + "__&@&__"
                + "UnservedQuantity"
                + "__&@&__"
                + str(elements[1])
            )
        ] = elements[5]
    return dic_of_parameterset


def transform_stock_data(data):
    # Splitting the first columns into 3 columns: TimeStep, StockId, Category
    # And adding the simulationRun
    temporary_df = data.copy()
    temporary_df.loc[:, "SimulationRun"] = EnvironmentVariables.simulation_id
    temporary_df[["StockId", "Category", "TimeStep"]] = temporary_df["index"].str.split(
        pat="__&@&__", expand=True
    )

    # Removing unused columns such as index, quantile 10 ...
    temporary_df = temporary_df.iloc[:, [1, 3, 5, 9, 14, 19, 23, 24, 25, 26, 27]]

    # Changing the columns' order and renaming them to match the adx table
    df_final = temporary_df[
        [
            "TimeStep",
            "SimulationRun",
            "StockId",
            "quantile 5%",
            "quantile 25%",
            "quantile 50%",
            "quantile 75%",
            "quantile 95%",
            "mean",
            "sem",
            "Category",
        ]
    ]

    df_final.rename(
        columns={
            "quantile 5%": "Percentile5",
            "quantile 25%": "Percentile25",
            "quantile 50%": "Percentile50",
            "quantile 75%": "Percentile75",
            "quantile 95%": "Percentile95",
            "mean": "Mean",
            "sem": "SE",
        },
        inplace=True,
    )
    return df_final


def transform_transport_data(data):

    # Adding the simulation run to the df
    temporary_df = data.copy()
    if not temporary_df.empty:
        temporary_df.loc[:, "SimulationRun"] = EnvironmentVariables.simulation_id
        # Splitting the first column into 2 columns: TransportOperation, TimeStep
        temporary_df[["TransportOperation", "TimeStep"]] = temporary_df[
            "index"
        ].str.split(pat="__&@&__", expand=True)
    else:
        temporary_df[["TransportOperation", "TimeStep"]] = None
    df_final = temporary_df.iloc[:, 1:]

    df_final.rename(
        columns={
            "confidence interval of the mean at 95%": "ConfidenceIntervalOfTheMeanAt95",
            "quantile 5%": "Percentile5",
            "quantile 10%": "Percentile10",
            "quantile 15%": "Percentile15",
            "quantile 20%": "Percentile20",
            "quantile 25%": "Percentile25",
            "quantile 30%": "Percentile30",
            "quantile 35%": "Percentile35",
            "quantile 40%": "Percentile40",
            "quantile 45%": "Percentile45",
            "quantile 50%": "Percentile50",
            "quantile 55%": "Percentile55",
            "quantile 60%": "Percentile60",
            "quantile 65%": "Percentile65",
            "quantile 70%": "Percentile70",
            "quantile 75%": "Percentile75",
            "quantile 80%": "Percentile80",
            "quantile 85%": "Percentile85",
            "quantile 90%": "Percentile90",
            "quantile 95%": "Percentile95",
        },
        inplace=True,
    )

    return df_final



import json
import os


def write_simulator_input(simulator_input: dict,
                          file_path: str = 'simulator_input.json') -> None:
    head, tail = os.path.split(file_path)
    if head:
        os.makedirs(head, exist_ok=True)
    if not tail.endswith('.json'):
        tail = f'{tail}.json'
    with open(os.path.join(head, tail), 'w') as f:
        f.write(json.dumps(simulator_input))


def read_simulator_input(file_path: str = 'simulator_input.json') -> dict:
    if not os.path.isfile(file_path) and not file_path.endswith('.json'):
        file_path = f'{file_path}.json'
    with open(file_path) as f:
        line = next(f)
    return json.loads(line)



class ValidationSchemas:
    schemas = {
        "Configuration": {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "properties": {
                "StartingDate": {"type": "string", "format": "date"},
                "SimulatedCycles": {"type": "integer", "minimum": 1},
                "StepsPerCycle": {"type": "integer", "minimum": 1},
                "TimeStepDuration": {"type": "integer", "minimum": 1},
                "ManageBacklogQuantities": {"type": "boolean"},
                "OptimizationObjective": {
                    "type": "string",
                    "enum": ["ServiceLevelMaximization", "ProfitMaximization"],
                },
                "ActivateUncertainties": {"type": "boolean"},
                "UncertaintiesProbabilityDistribution": {
                    "type": "string",
                    "enum": ["Uniform", "Gaussian"],
                },
                "TransportUncertaintiesProbabilityDistribution": {
                    "type": "string",
                    "enum": [
                        "alpha",
                        "arcsine",
                        "beta",
                        "betaprime",
                        "burr",
                        "burr12",
                        "cauchy",
                        "chi",
                        "chi2",
                        "cosine",
                        "dgamma",
                        "dweibull",
                        "exponential",
                        "exponnorm",
                        "exponweib",
                        "f",
                        "fatiguelife",
                        "fisk",
                        "foldnorm",
                        "gamma",
                        "gengamma",
                        "genlogistic",
                        "gennorm",
                        "genexpon",
                        "genextreme",
                        "gilbrat",
                        "gompertz",
                        "halfcauchy",
                        "halfnorm",
                        "invgamma",
                        "invgauss",
                        "invweibull",
                        "laplace",
                        "logistic",
                        "loggamma",
                        "lognormal",
                        "loguniform",
                        "lomax",
                        "normal",
                        "pareto",
                        "powerlaw",
                        "skewnorm",
                        "t",
                        "trapezoid",
                        "triangular",
                        "truncexpon",
                        "truncnorm",
                        "uniform",
                        "vonmises",
                        "weibull",
                        "bernoulli",
                        "betabinom",
                        "binomial",
                        "dlaplace",
                        "discreteuniform",
                        "geom",
                        "hypergeom",
                        "logser",
                        "poisson",
                    ],
                },
                "FinancialCostOfStock": {"type": "number", "minimum": 0, "maximum": 1},
                "BatchSize": {"type": "integer", "minimum": 0},
                "EmptyObsoleteStocks": {"type": "boolean"},
                "ActivateVariableMachineOpeningRate": {"type": "boolean"},
                "EnforceProductionPlan": {"type": "boolean"},
                "IntermediaryStockDispatchPolicy": {
                    "type": "string",
                    "enum": ["DispatchAll", "AllowRetention"],
                },
                "ActualizeShipments": {"type": "boolean"},
                "ActivateCorrelatedDemandUncertainties": {"type": "boolean"},
                "DemandCorrelations": {"type": "number", "minimum": 0, "maximum": 1},
            },
        },
        "input": {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "properties": {"InputQuantity": {"type": "number", "exclusiveMinimum": 0}},
        },
        "ProductionOperation": {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "properties": {
                "QuantitiesToProduce": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "OperatingPerformances": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {"type": "number", "minimum": 0, "maximum": 1}
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "CycleTimes": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "RejectRates": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {"type": "number", "minimum": 0, "maximum": 1}
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "OperatingPerformanceUncertainties": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {"type": "number", "minimum": 0, "maximum": 1}
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "ProductionUnitCosts": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "CO2UnitEmissions": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "MinimumOrderQuantities": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "MultipleOrderQuantities": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "SourcingProportions": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0, "maximum": 1}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "IsContractor": {"type": "boolean"},
                "InvestmentCost": {
                    "type": "number",
                    "minimum": 0,
                },
                "Priority": {"type": "integer", "minimum": 0},
                "Duration": {"type": "integer", "minimum": 0},
            },
            "required": ["CycleTimes"],
            "dependencies": {
                "OperatingPerformanceUncertainties": ["OperatingPerformances"]
            },
        },
        "ProductionResource": {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "properties": {
                "ProductionPolicy": {
                    "type": "string",
                    "enum": [
                        "None",
                        "Equidistribution",
                        "GreatestWorkload",
                        "SmallestWorkload",
                        "HighestPriority",
                    ],
                },
                "OpeningTimes": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "FixedProductionCosts": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "Latitude": {"type": "number", "minimum": -90, "maximum": 90},
                "Longitude": {"type": "number", "minimum": -180, "maximum": 180},
            },
            "required": ["OpeningTimes"],
        },
        "Stock": {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "properties": {
                "StorageCosts": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "IsInfinite": {"type": "boolean"},
                "MinimalStock": {
                    "type": "number",
                    "if": {"maximum": -1},
                    "then": {"minimum": -1},
                    "else": {"minimum": 0},
                },
                "MaximalStock": {
                    "type": "number",
                    "if": {"maximum": -1},
                    "then": {"minimum": -1},
                    "else": {"minimum": 0},
                },
                "InitialStock": {"type": "number", "minimum": 0},
                "InitialValue": {"type": "number", "minimum": 0},
                "PurchasingUnitCosts": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "CO2UnitEmissions": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "UnitIncomes": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "Demands": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                },
                "DemandUncertainties": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {"type": "number", "minimum": 0, "maximum": 1}
                    },
                },
                "DemandWeights": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                },
                "BacklogWeight": {"type": "number", "minimum": 0},
                "MaximizationWeight": {"type": "number", "minimum": 0},
                "StockPolicy": {
                    "type": "string",
                    "enum": [
                        "None",
                        "OrderPointFixedQuantity",
                        "OrderPointVariableQuantity",
                        "MakeToForecast",
                    ],
                },
                "SourcingPolicy": {
                    "type": "string",
                    "enum": [
                        "Equidistribution",
                        "HighestStock",
                        "HighestPriority",
                        "SourcingProportions",
                    ],
                },
                "DispatchPolicy": {
                    "type": "string",
                    "enum": [
                        "None",
                        "Equidistribution",
                        "GreatestQuantity",
                        "SmallestQuantity",
                        "HighestPriority",
                    ],
                },
                "ReviewPeriod": {"type": "integer", "minimum": 1},
                "FirstReview": {"type": "integer", "minimum": 0},
                "OrderPoints": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "OrderQuantities": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "OrderUpToLevels": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "Advance": {"type": "integer", "minimum": 0},
                "SafetyQuantities": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "SalesForecasts": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                },
                "Latitude": {"type": "number", "minimum": -90, "maximum": 90},
                "Longitude": {"type": "number", "minimum": -180, "maximum": 180},
            },
            "dependencies": {"InitialValue": ["InitialStock"]},
        },
        "Transport": {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "properties": {
                "Duration": {"type": "integer", "minimum": 0},
                "Priority": {"type": "integer", "minimum": 0},
                "Mode": {"type": "string"},
                "InitialTransportedQuantities": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                },
                "InitialTransportedValues": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                },
                "CustomFees": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "TransportUnitCosts": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "CO2UnitEmissions": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "ActualDurations": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {"type": "integer", "minimum": 0}
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "MinimumOrderQuantities": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "MultipleOrderQuantities": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "SourcingProportions": {
                    "type": "object",
                    "patternProperties": {"^[0-9]+$": {"type": "number", "minimum": 0, "maximum": 1}},
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "TransportUncertaintiesParameter1": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                        }
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "TransportUncertaintiesParameter2": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                        }
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "TransportUncertaintiesParameter3": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                        }
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
                "TransportUncertaintiesParameter4": {
                    "type": "object",
                    "patternProperties": {
                        "^[0-9]+$": {
                            "type": "number",
                        }
                    },
                    "if": {"minProperties": 1},
                    "then": {"required": ["0"]},
                },
            },
            "required": ["Duration"],
            "dependencies": {
                "InitialTransportedValues": ["InitialTransportedQuantities"]
            },
        },
    }
    graph = {
        "contains": {
            "source": "source",
            "target": "target",
            "links": {
                "source": ["ProductionResource", "id"],
                "target": ["ProductionOperation", "id"],
            },
            "cardinalities": "1:N",
            "all_target_present": True,
            "all_source_present": True,
        },
        "input": {
            "source": "source",
            "target": "target",
            "links": {
                "source": ["Stock", "id"],
                "target": ["ProductionOperation", "id"],
            },
            "cardinalities": "N:N",
            "all_target_present": True,
            "all_source_present": False,
        },
        "output": {
            "source": "source",
            "target": "target",
            "links": {
                "source": ["ProductionOperation", "id"],
                "target": ["Stock", "id"],
            },
            "cardinalities": "N:1",
            "all_target_present": False,
            "all_source_present": True,
        },
        "Transport": {
            "source": "source",
            "target": "target",
            "links": {"source": ["Stock", "id"], "target": ["Stock", "id"]},
            "cardinalities": "N:N",
            "all_target_present": False,
            "all_source_present": False,
        },
    }



def seconds_to_iso_8061_duration(total_time_in_seconds: int) -> str:
    """
    Convert a duration in seconds to an iso 8061 formated string
    :param total_time_in_seconds: duration to convert
    :return: iso 8061 representation of the duration
    """
    time_left = total_time_in_seconds

    minute_duration = 60
    hour_duration = minute_duration * 60
    day_duration = hour_duration * 24

    days = time_left // day_duration
    time_left -= days * day_duration
    hours = time_left // hour_duration
    time_left -= hours * hour_duration
    minutes = time_left // minute_duration
    time_left -= minutes * minute_duration

    return f'P{days}DT{hours}H{minutes}M{time_left}S'



from typing import Union

from Supplychain.Wrappers.simulator import CosmoEngine
from Supplychain.Generic.adx_and_file_writer import ADXAndFileWriter
import pandas as pd
from Supplychain.Generic.timer import Timer
from Supplychain.Wrappers.environment_variables import EnvironmentVariables
from Supplychain.Run.simulation import run_simple_simulation
import comets as co

# parameter label: (entity_type, attribute_name, default)
sensitive_parameters = {
    "Fixed Production Cost": (
        "Machine",
        "ProductionCostSchedule",
        "ProductionCost",
    ),
    "Production Resource Opening Time": (
        "Machine",
        "OpeningTimeSchedule",
        "OpeningTime",
    ),
    "Operating Performance": (
        "Operation",
        "OperatingPerformanceSchedule",
        "OperatingPerformance",
    ),
    "Cycle Time": (
        "Operation",
        "CycleTimeSchedule",
        "CycleTime",
    ),
    "Variable Production Cost": (
        "Operation",
        "ProductionUnitCostSchedule",
        "ProductionUnitCost",
    ),
    "Production CO2 Unit Emissions": (
        "Operation",
        "CO2UnitEmissionsSchedule",
        "CO2UnitEmissions",
    ),
    "Production Minimum Order Quantity": (
        "Operation",
        "MinimumOrderQuantitySchedule",
        "MinimumOrderQuantity",
    ),
    "Production Multiple Order Quantity": (
        "Operation",
        "MultipleOrderQuantitySchedule",
        "MultipleOrderQuantity",
    ),
    "Production Plan": (
        "Operation",
        "ProductionPlanSchedule",
        "ProductionPlan",
    ),
    "Initial Stock": (
        "Stock",
        "CurrentStock",
        None,
    ),
    "Purchasing Unit Cost": (
        "Stock",
        "PurchasingUnitCostSchedule",
        "PurchasingUnitCost",
    ),
    "Unit Income": (
        "Stock",
        "UnitIncomeSchedule",
        "UnitIncome",
    ),
    "Storage Unit Cost": (
        "Stock",
        "StorageUnitCostSchedule",
        "StorageUnitCost",
    ),
    "Order Point": (
        "Stock",
        "OrderPointSchedule",
        "OrderPoint",
    ),
    "Order Quantities": (
        "Stock",
        "OrderQuantitySchedule",
        "OrderQuantity",
    ),
    "Order Up To Levels": (
        "Stock",
        "OrderQuantitySchedule",
        "OrderQuantity",
    ),
    "Safety Quantities": (
        "Stock",
        "SafetyQuantitySchedule",
        "SafetyQuantity",
    ),
    "Transport Unit Cost": (
        "TransportOperation",
        "TransportUnitCostSchedule",
        "TransportUnitCost",
    ),
    "Custom Fees": (
        "TransportOperation",
        "CustomsUnitCostSchedule",
        "CustomsUnitCost",
    ),
    "Transport Duration": (
        "TransportOperation",
        "ActualDurationSchedule",
        "Duration",
    ),
    "Transport CO2 Unit Emission": (
        "TransportOperation",
        "CO2UnitEmissionsSchedule",
        "CO2UnitEmissions",
    ),
    "Transport Minimum Order Quantity": (
        "TransportOperation",
        "MinimumOrderQuantitySchedule",
        "MinimumOrderQuantity",
    ),
    "Transport Multiple Order Quantity": (
        "TransportOperation",
        "MultipleOrderQuantitySchedule",
        "MultipleOrderQuantity",
    ),
}

# For each parameter of the sensitivity analysis, to which entities it corresponds to
map_to_entities = {
    label: entity_type
    for label, (entity_type, _, _)
    in sensitive_parameters.items()
}

# For each parameter of the sensitivity analysis, to which attribute of the model it corresponds to
map_to_attribute_name = {
    label: attribute_name
    for label, (_, attribute_name, _)
    in sensitive_parameters.items()
}

# For each parameter of the sensitivity analysis, which attribute of the model
# should be used as default value to create a schedule if the provided schedule is empty
map_to_default_attribute_name = {
    label: default
    for label, (_, _, default)
    in sensitive_parameters.items()
}


def local_sensitivity_analysis(
    simulation_name: str,
    amqp_consumer_adress: Union[str, None] = None,
    parameter: str = "Transport Duration",  # type of variable on which the analysis is performed
    adx_writer: Union[ADXAndFileWriter, None] = None,
    variation: float = 0.5,  # variation value
    change: str = "relative",  # change mode
    local: bool = False,  # false to optionally send the results to adx ; true to return the results dataframe (useful for the test)
    simulator_path: str = "Simulation",
    saving_results: bool = False,
    validation_folder: str = "Simulation/Output",  # Folder in which sensitivity analysis results are saved
    timeinterval: bool = False,  # Whether the change happens during a specific interval (otherwise it is applied for the complete duration of the simulation)
    initialtimestep: int = 0,  # Time step at which the modification starts to take place (included), only if timeinterval is true
    finaltimestep: int = 0,  # Time step after which the modification ends (the modification still takes place at this time step), only if timeinterval is true
):

    with Timer("[Run Sensitivity]") as t:

        if parameter != "All":
            results = run_one_sa(
                simulation_name,
                amqp_consumer_adress,
                parameter,
                adx_writer,
                variation,
                change,
                local,
                simulator_path,
                saving_results,
                validation_folder,
                timeinterval,
                initialtimestep,
                finaltimestep,
            )
        else:
            # Run a sensitivity analysis for each possible parameter and append results to the same table

            result_dataframes = []
            for single_parameter in map_to_entities.keys():
                try:
                    t.display_message(
                        f"Running sensitivity analysis with parameter {single_parameter}"
                    )
                    results = run_one_sa(
                        simulation_name,
                        amqp_consumer_adress,
                        single_parameter,
                        adx_writer,
                        variation,
                        change,
                        local,
                        simulator_path,
                        saving_results,
                        validation_folder,
                        timeinterval,
                        initialtimestep,
                        finaltimestep,
                    )
                    result_dataframes.append(results)
                except Exception as e:
                    t.display_message(
                        f"Error running sensitivity analysis with parameter {single_parameter}, analysis skipped"
                    )
                    print(e)
            results = pd.concat(result_dataframes)

        if saving_results:
            results.to_csv(validation_folder + "/SensitivityAnalysis.csv")

        # useful for tests
        if local:
            return results

        else:
            adx_writer.write_target_file(
                results.to_dict("records"),
                "LocalSensitivityAnalysis",
                EnvironmentVariables.simulation_id,
            )

            t.split("Sent stats to ADX : {time_since_last_split}")
            t.display_message("Running simple simulation to fill ADX")

            run_simple_simulation(
                simulation_name=simulation_name,
                amqp_consumer_adress=amqp_consumer_adress,
            )


def run_one_sa(
    simulation_name: str,
    amqp_consumer_adress: Union[str, None] = None,
    parameter: str = "Transport Duration",  # type of variable on which the analysis is performed
    adx_writer: Union[ADXAndFileWriter, None] = None,
    variation: float = 0.5,  # variation value
    change: str = "relative",  # change mode
    local: bool = False,  # false to optionally send the results to adx ; true to return the results dataframe (useful for the test)
    simulator_path: str = "Simulation",
    saving_results: bool = False,
    validation_folder: str = "Simulation/Output",  # Folder in which sensitivity analysis results are saved
    timeinterval: bool = False,  # Whether the change happens during a specific interval (otherwise it is applied for the complete duration of the simulation)
    initialtimestep: int = 0,  # Time step at which the modification starts to take place (included), only if timeinterval is true
    finaltimestep: int = 0,  # Time step after which the modification ends (the modification still takes place at this time step), only if timeinterval is true
):

    with Timer("[Run one sensitivity analysis]") as t:
        attribute = map_to_attribute_name.get(parameter)

        list_of_datapaths, max_time_step = get_simulator_data(
            simulator_path, map_to_entities.get(parameter), attribute
        )

        errors = validate_sa_parameters(
            parameter,
            change,
            variation,
            timeinterval,
            initialtimestep,
            finaltimestep,
            max_time_step,
        )
        if errors:
            t.display_message('\n'.join(errors))
            error_message = f"Invalid parameter{'s' if len(errors) > 1 else ''}"
            raise ValueError(error_message)

        list_of_variables = create_variables(
            simulator_path,
            list_of_datapaths,
            change,
            variation,
            timeinterval,
            initialtimestep,
            finaltimestep,
            max_time_step,
            parameter,
            attribute,
        )

        errors = validate_absolute_variations(list_of_variables)
        if errors:
            t.display_message('\n'.join(errors))
            error_message = f"Invalid parameter{'s' if len(errors) > 1 else ''}"
            raise ValueError(error_message)

        def encoder(parameters):
            """
            Encoder of the task on which the S.A will be performed. The encoder receives a parameter set where
            the attributes of type 'schedule' are separated (see function create_variables). The aim of this encoder is to
            re concatenate the schedules together. For example, if the encoder receives the following parameter set:
            {"transport_lyon": 2,
                "transport_paris__@__0": 1,
                "transport_paris__@__3": 2,
                "transport_paris__@__5": 4,}

            it will return :
            {'transport_lyon': 2,
                'transport_paris': {'0': 1, '3': 2, '5': 4}}
            Args:
                parameters : Parameter set with the value computed by the S.A algorithm where schedules are separated.
                Note that the keys of the parameter set are either a datatapth or a datapath + '__@__' + time_step.

            Returns:
                parameter_set : Parameter set with the value computed by the S.A algorithm where the schedules are re concatenated.
                Note that the keys of the parameter set are the datapath of the attributes.
            """
            parameter_set = {}
            for key, value in parameters.items():
                if "__@__" in key:  # check if it belongs to a group
                    if key.split("__@__")[0] not in parameter_set.keys():
                        parameter_set[key.split("__@__")[0]] = {
                            key.split("__@__")[-1]: value
                        }
                    else:
                        temp_dic = parameter_set[key.split("__@__")[0]].copy()
                        temp_dic[key.split("__@__")[-1]] = value
                        parameter_set[key.split("__@__")[0]] = temp_dic
                else:
                    parameter_set[key] = value
            return parameter_set

        used_probes = ["PerformanceIndicators"]

        class PerformanceConsumer:
            def __init__(self):
                self.data = {}

            def Consume(self, p_data):
                f = self.engine.PerformanceIndicatorsProbeOutput.Cast(p_data).GetFacts()
                for fact in f:
                    self.data = {
                        "OPEX": fact.GetOPEX().GetAsFloat(),
                        "Profit": fact.GetProfit().GetAsFloat(),
                        "Average stock value": fact.GetAverageStockValue().GetAsFloat(),
                        "CO2 emissions": fact.GetCO2Emissions().GetAsFloat(),
                        "Total served quantity": fact.GetTotalServedQuantity().GetAsFloat(),
                        "Fill rate service level": (
                            fact.GetTotalServedQuantity().GetAsFloat() / fact.GetTotalDemand().GetAsFloat() * 100
                            if fact.GetTotalDemand().GetAsFloat() != 0
                            else 0
                        ),
                    }

        simulator_interface = co.CosmoInterface(
            simulator_path=simulator_path,
            custom_sim_engine=CosmoEngine,
            simulation_name=simulation_name,
            used_consumers=[],
            used_probes=used_probes,
            custom_consumers=[
                (PerformanceConsumer, "PerformanceConsumer", "PerformanceIndicators")
            ],
        )

        def get_outcomes(simulator_interface):
            data = simulator_interface.PerformanceConsumer.data
            return data

        simulationtask = co.ModelTask(
            modelinterface=simulator_interface,
            get_outcomes=get_outcomes,
            encode=encoder,
        )

        experiment = co.LocalSensitivityAnalysis(
            task=simulationtask, variables=list_of_variables, n_jobs=-2
        )

        t.display_message("Starting simulations")
        experiment.run()
        t.split("Ended simulations : {time_since_start}")

        results = reformat_results(
            experiment.results,
            simulation_name,
            change,
            variation,
            parameter,
            timeinterval,
            initialtimestep,
            finaltimestep,
        )

    return results


def validate_sa_parameters(
    sensitive_parameter: str,
    change: str,
    variation: float,
    timeinterval: bool,
    initialtimestep: int,
    finaltimestep: int,
    max_time_step: int,
):
    errors = []

    for label, parameter, parameter_type in [
        ("sensitive parameter", sensitive_parameter, str),
        ("change", change, str),
        ("variation", variation, float),
        ("time interval", timeinterval, bool),
        ("initial time step", initialtimestep, int),
        ("final time step", finaltimestep, int),
    ]:
        if not isinstance(parameter, parameter_type):
            errors.append(
                f"Invalid {label} value: {parameter} is not {parameter_type.__name__}."
            )

    known_parameter = True
    if sensitive_parameter not in sensitive_parameters:
        known_parameter = False
        errors.append(
            f"Unkown sensitive parameter: {sensitive_parameter}. Parameter should be in {sensitive_parameters.keys()}."
        )

    changes = set(("relative", "absolute", "replacement"))
    if change not in changes:
        errors.append(
            f"Unknown change: {change}, it should be in {changes}."
        )
    elif change == "relative" and variation < -1:
        errors.append(
            f"Invalid variation value: {variation}. With relative change, it should be greater than or equal to -1."
        )
    elif change == "replacement" and variation < 0:
        errors.append(
            f"Invalid variation value: {variation}. With replacement change, it should be non negative."
        )

    if timeinterval:
        if known_parameter and map_to_default_attribute_name[sensitive_parameter] is None:
            errors.append(
                f"Unsupported sensitivity analysis: as {sensitive_parameter} cannot be scheduled, time interval should not be activated."
            )
        if (
            initialtimestep > finaltimestep
            or initialtimestep < 0
            or initialtimestep > max_time_step
            or finaltimestep < 0
            or finaltimestep > max_time_step
        ):
            errors.append(
                f"Time steps {initialtimestep} and {finaltimestep} should be integers between 0 and {max_time_step}"
            )

    return errors


def validate_absolute_variations(variables):
    errors = []

    for variable in variables:
        if variable["change"] == "absolute" and variable["variation"] < -variable["reference"]:
            entity_name = variable["name"].split('::{Entity}')[-1].split('::@')[0]
            time_step = variable["name"].split('__@__')[-1]
            errors.append(
                f"Invalid variation value: {variable['variation']}."
                f" With absolute change, it should be greater than or equal to {-variable['reference']}"
                f" for entity {entity_name} at time step {time_step}."
            )

    return errors


def get_simulator_data(simulator_path, entity_type, attribute):
    """
    Function to automatically retrieve the list of datapath on which the S.A will be performed, based on the entity type
    specified by the user.

    Args:
        simulator_path : path to the simulator
        entity_type : type of the entities on which the S.A will be performed
    Returns:
        (list, integer): list of datapaths on which the S.A will be used and number of time steps of the simulation
    """
    sim2 = co.CosmoInterface(simulator_path, custom_sim_engine=CosmoEngine)
    sim2.initialize()

    list_of_datapaths = (
        get_list_of_datapaths(sim2, entity_type, attribute)
        if entity_type and attribute
        else None
    )

    max_time_step = get_max_time_step(sim2)

    sim2.terminate()

    return list_of_datapaths, max_time_step


def get_list_of_datapaths(cosmo_interface, entity_type, attribute):
    suffixes = sorted(
        f"{{Entity}}{entity.GetName()}::@{attribute}"
        for entity in cosmo_interface.sim.GetModel().FindEntitiesByType(entity_type)
    )

    if not suffixes:
        raise ValueError(
            f"No entity of type {entity_type} is present in the instance, the sensitivity analysis on {attribute} is not supported."
        )

    datapaths = cosmo_interface.get_datapaths()

    return [
        next(filter(
            lambda datapath: datapath.endswith(suffix),
            datapaths,
        ))
        for suffix in suffixes
    ]


def get_max_time_step(cosmo_interface):
    p = "Model::@TimeStepPerCycle"
    time_step_per_cycle = cosmo_interface.get_outputs([p])[p]
    p = "Model::@NumberOfCycle"
    number_of_cycle = cosmo_interface.get_outputs([p])[p]
    return time_step_per_cycle * number_of_cycle


def create_variables(
    simulator_path,
    datapaths,
    change,
    variation,
    timeinterval,
    initialtimestep,
    finaltimestep,
    max_time_step,
    parameter,
    attribute,
):
    """
    Function to create the list of variables to perform the S.A based on a list of datapaths.
    Each datapath corresponds to an attribute in the model. If the attribute is of type "schedule":
    i.e {'time_step_0': value on time step 0, [...], 'time_step_k': value on time step k}
    the function will create more than one variable per datapath. More precisely n variables all belonging to
    the same group will be created, where n corresponds to the number of items in the dic. For example,
    the following attribute : {'0': 3, '2': 4, '7':2} will lead to the creation of the following variables:
    {"name": datapath__@__0,
     "type": int),
     "reference": 3,
     "variation": variation,
     "change": change,
     "group": datapath,
                }
    {"name": datapath__@__2,
     "type": int),
     "reference": 4,
     "variation": variation,
     "change": change,
     "group": datapath,
                }
    {"name": datapath__@__7,
     "type": int),
     "reference": 2,
     "variation": variation,
     "change": change,
     "group": datapath,
                }
    In the case of timeinterval==true, a variable will also be created for time steps initialtimestep and finaltimestep+1,
    even if they were not already present in the schedule, in order to start the change of value, and to remove it.

    Args:
        simulator_path : path to the simulator
        datapaths : list containing the datapaths that need to be mapped to a variable
        change : change mode of the S.A
        variation : variation of the variable for the S.A

    Returns
        list_of_variables : list of dic where each dic is a variable according to the
        definition of a variable in CoMETs' sensitivity analysis

    """
    sim = CosmoEngine.LoadSimulator(simulator_path)
    list_of_variables = []
    for elements in datapaths:
        reference = sim.FindAttribute(elements).Get()

        if not timeinterval:
            if isinstance(reference, dict):

                if reference == {}:
                    if map_to_default_attribute_name[parameter] is not None:
                        newelement = elements.replace(
                            map_to_attribute_name[parameter],
                            map_to_default_attribute_name[parameter],
                        )
                        new_reference = sim.FindAttribute(newelement).Get()
                        reference = {0: new_reference}
                    else:
                        raise ValueError(
                            f"Schedule for attribute {elements} is empty and no default value is available, sensitivity analysis requires a reference value to be provided."
                        )

                for key in reference:
                    variable = {
                        "name": str(elements + "__@__" + str(key)),
                        "type": get_type(reference[key]),
                        "reference": reference[key],
                        "variation": variation,
                        "change": change,
                        "group": elements,
                    }
                    list_of_variables.append(variable)

            else:
                variable = {
                    "name": elements,
                    "type": get_type(reference),
                    "reference": reference,
                    "variation": variation,
                    "change": change,
                }
                list_of_variables.append(variable)
        else:

            if not isinstance(reference, dict):
                raise ValueError(
                    f"No schedule is provided for attribute {elements}, sensitivity analysis with timeinterval is not supported."
                )

            if reference == {}:
                if map_to_default_attribute_name[parameter] is not None:
                    newelement = elements.replace(
                        map_to_attribute_name[parameter],
                        map_to_default_attribute_name[parameter],
                    )
                    new_reference = sim.FindAttribute(newelement).Get()
                    reference = {0: new_reference}
                else:
                    raise ValueError(
                        f"Schedule for attribute {elements} is empty and no default value is available, sensitivity analysis requires a reference value to be provided."
                    )

            previousvalue = 0  # keep track of last defined value in the schedule
            for timestep in range(max_time_step):
                if timestep < initialtimestep or (timestep > finaltimestep + 1):
                    # outside the interval, no variation should be applied
                    if timestep in reference.keys():
                        variable = {
                            "name": str(elements + "__@__" + str(timestep)),
                            "type": get_type(reference[timestep]),
                            "reference": reference[timestep],
                            "variation": 0,
                            "change": "absolute",
                            "group": elements,
                        }
                        list_of_variables.append(variable)
                        previousvalue = reference[timestep]
                elif timestep == initialtimestep:
                    # start the change of value at initialtimestep,
                    # if initialtimestep is not in the schedule, use the last defined value previousvalue as reference
                    if timestep in reference.keys():
                        variable = {
                            "name": str(elements + "__@__" + str(timestep)),
                            "type": get_type(reference[timestep]),
                            "reference": reference[timestep],
                            "variation": variation,
                            "change": change,
                            "group": elements,
                        }
                        list_of_variables.append(variable)
                        previousvalue = reference[timestep]
                    else:
                        variable = {
                            "name": str(elements + "__@__" + str(timestep)),
                            "type": get_type(previousvalue),
                            "reference": previousvalue,
                            "variation": variation,
                            "change": change,
                            "group": elements,
                        }
                        list_of_variables.append(variable)
                elif timestep > initialtimestep and timestep <= finaltimestep:
                    # inside the interval, the variation should be applied
                    if timestep in reference.keys():
                        variable = {
                            "name": str(elements + "__@__" + str(timestep)),
                            "type": get_type(reference[timestep]),
                            "reference": reference[timestep],
                            "variation": variation,
                            "change": change,
                            "group": elements,
                        }
                        list_of_variables.append(variable)
                        previousvalue = reference[timestep]
                elif timestep == finaltimestep + 1:
                    # end the change of value at finaltimestep+1,
                    # if finaltimestep is not in the schedule, use the last defined value previousvalue as reference
                    if timestep in reference.keys():
                        variable = {
                            "name": str(elements + "__@__" + str(timestep)),
                            "type": get_type(reference[timestep]),
                            "reference": reference[timestep],
                            "variation": 0,
                            "change": "absolute",
                            "group": elements,
                        }
                        list_of_variables.append(variable)
                        previousvalue = reference[timestep]
                    else:
                        variable = {
                            "name": str(elements + "__@__" + str(timestep)),
                            "type": get_type(previousvalue),
                            "reference": previousvalue,
                            "variation": 0,
                            "change": "absolute",
                            "group": elements,
                        }
                        list_of_variables.append(variable)
    return list_of_variables


def get_type(value):
    """
    Function to return the value's type. Instead of returning "<class 'int'>" the function returns 'int'.
    Similarly, for float, instead of <class 'float'>" the function returns 'float'
    """
    return type(value).__name__


def reformat_results(
    row_results,
    simulation_name,
    change,
    variation,
    parameter,
    timeinterval,
    initialtimestep,
    finaltimestep,
):
    """
    Function to reformat the S.A results to be compatible with the adx table.

    """
    df = row_results.reset_index()

    if (
        "Group" not in df.columns
    ):  # if there is no 'Group' column, creates one where the name of the groups correspond to the ID.
        df.insert(1, "Group", df["Name"])

    df["Name"] = df["Name"].apply(
        lambda name: (
            name.split("{Entity}")[-1].split("::@", 1)[0]
            + (f"__@__{name.rsplit('__@__', 1)[1]}" if "__@__" in name else "__@__0")
        )
    )
    df[["Name", "TimeStep"]] = df["Name"].str.split("__@__", n=1, expand=True)
    df["TimeStep"] = df["TimeStep"].astype(int)

    # Filter lines for time steps where no modification has been done
    if timeinterval:
        df = df[(df["TimeStep"] >= initialtimestep) & (df["TimeStep"] <= finaltimestep)]

    df["SimulationRun"] = EnvironmentVariables.simulation_id
    df["ScenarioName"] = simulation_name
    df["Variable"] = str(parameter)
    df["Variation"] = float(variation)
    df["Change"] = str(change)
    df["TimeInterval"] = bool(timeinterval)
    df["InitialTimeStep"] = int(initialtimestep)
    df["FinalTimeStep"] = int(finaltimestep)

    df = df.rename(
        columns={
            "Output": "KPI",
            "Name": "ID",
            "ReferenceOutputValue": "ReferenceKPI",
            "NewOutputValue": "NewKPI",
            "Difference": "Gap",
        }
    )
    # if type(df["ReferenceInputValue"][0]) == dict:
    #     for i in range(len(df.index)):
    #         df["ReferenceInputValue"][i] = df["ReferenceInputValue"][i][0]
    #         df["NewInputValue"][i] = df["NewInputValue"][i][0]

    return df



from Supplychain.Generic.timer import Timer
import os
import uuid

from azure.identity import DefaultAzureCredential
from azure.digitaltwins.core import DigitalTwinsClient


class ADTWriter(Timer):

    def send_twin(self, twin: dict) -> dict:
        digital_twin_client = DigitalTwinsClient(os.environ["AZURE_DIGITAL_TWINS_URL"], self.azure_credentials)
        new_item = dict()
        for key in twin.keys():
            if type(twin[key]) is not dict or twin[key]:
                new_item[key] = twin[key]
        new_item.setdefault('$id', str(uuid.uuid1()))
        item_id = twin['$id'].replace(" ", "")
        return digital_twin_client.upsert_digital_twin(item_id,
                                                       new_item)

    def send_relation(self, relation: dict) -> dict:
        digital_twin_client = DigitalTwinsClient(os.environ["AZURE_DIGITAL_TWINS_URL"], self.azure_credentials)
        relation['$sourceId'] = relation['$sourceId'].replace(" ", "")
        relation['$targetId'] = relation['$targetId'].replace(" ", "")
        relation.setdefault('$relationshipId', str(uuid.uuid1()))
        relation['$relationshipId'] = relation['$relationshipId'].replace(" ", "")
        new_item = dict()
        for key in relation.keys():
            if type(relation[key]) is not dict or relation[key]:
                new_item[key] = relation[key]
        return digital_twin_client.upsert_relationship(relation['$sourceId'],
                                                       relation['$relationshipId'],
                                                       new_item)

    def send_items(self,
                   items: list):
        self.reset()
        for item in items:
            _ = self.send_twin(item) if '$sourceId' not in item else self.send_relation(item)
        self.display_message(f"Sent {len(items)} items in " + "{time_since_start:6.4f}s")

    def purge_adt(self, query: str = 'SELECT * FROM digitaltwins', delete_relation: bool = True):
        digital_twin_client = DigitalTwinsClient(os.environ["AZURE_DIGITAL_TWINS_URL"], self.azure_credentials)
        self.display_message("Querying twins")
        twin_list = []
        for item in digital_twin_client.query_twins(query):
            twin_id = str(item['$dtId'])
            twin_list.append(twin_id)
            current_length = len(twin_list)
            if current_length % 100 == 0:
                self.display_message(f"Found {current_length} twins")
        self.display_message(f"Found a total of {len(twin_list)} twins")
        if delete_relation:
            self.display_message("Deleting relationships")
            for twin_id in twin_list:
                for relation in digital_twin_client.list_relationships(twin_id):
                    relation_id = relation['$relationshipId']
                    digital_twin_client.delete_relationship(twin_id, relation_id)
        self.display_message("Deleting twins")
        for twin_id in twin_list:
            digital_twin_client.delete_digital_twin(twin_id)
        self.display_message("Purge complete")

    def __init__(self, force_purge: bool = False):
        Timer.__init__(self, "[ADT Writer]")
        self.azure_credentials = DefaultAzureCredential()
        if force_purge:
            self.display_message("Forcing purge of ADT")
            self.purge_adt()



from csm.engine import SimulatorInterface
from Supplychain.Generic.adx_and_file_writer import ADXAndFileWriter
import pandas as pd


def send_production_route_data(writer: ADXAndFileWriter,
                               simulation_id: str,
                               simulator: SimulatorInterface):
    df = get_sorted_production_resource(simulator)
    df['SimulationRun'] = simulation_id
    dict_list = df.to_dict('records')
    writer.write_target_file(dict_list, 'ProductionRoute', simulation_id)


def get_operations_for_stocks(output_op, op_input, stocks):
    operations = set()
    
    for stock in stocks:
        if output_op.get(stock):
            operations.update(output_op.get(stock).get('Operation', {}))            
            transports = output_op.get(stock).get('TransportOperation', {})
            
            # Transports are ignored in the production route
            # If there are transports for the stocks, we will recursively research the previous operations
            if transports:
                previous_stocks = get_stocks_for_operations(op_input, transports)
                previous_operations = get_operations_for_stocks(output_op, op_input, previous_stocks)
                operations.update(previous_operations)
                
    return operations


def get_stocks_for_operations(op_input, operations):
    stocks = set()
    for operation in operations:
        stock = op_input.get(operation)
        if stock:
            stocks.update(set(stock[0]))
    return stocks


def get_machines_for_operations(op_machines, operations):
    machines = {}
    for op in operations:
        machine = op_machines.get(op)
        if machine:
            machines[op] = machine.GetName()
    return machines


def get_sorted_production_resource(simulator):
    # Return sorted Production Resource by Production Step Order by Stock
    stocks = [stock.GetName() for stock in simulator.get_entities_by_type('Stock')]
    op_input = simulator.get_stocks_by_operation()
    output_op = simulator.get_typed_operations_by_output_stock()
    op_machines = simulator.get_machines_by_operation()
    record = []

    for stock in stocks:
        operations = get_operations_for_stocks(output_op, op_input, [stock])
        machines_by_step = []

        while operations:
            machines = get_machines_for_operations(op_machines, operations)
            if machines:
                machines_by_step.append(machines)
            previous_input_stocks = get_stocks_for_operations(op_input, operations)
            previous_operations = get_operations_for_stocks(output_op, op_input, previous_input_stocks)
            operations = previous_operations

        machines_by_step.reverse()

        for i in range(len(machines_by_step)):
            for op, machine in machines_by_step[i].items():
                record.append({'ProductionStepOrder': i + 1, 'ProductionOperationId': op,
                               'ProductionResourceId': machine, 'StockId': stock})

        df = pd.DataFrame(record)
    return df



from Supplychain.Generic.folder_io import FolderReader, FolderWriter
from Supplychain.Generic.timer import Timer
from Supplychain.Schema.simulator_files_description import simulator_format
from Supplychain.Schema.default_values import variables_default_values
from Supplychain.Generic.duration import seconds_to_iso_8061_duration
from datetime import datetime, tzinfo, timedelta
from Supplychain.Wrappers.environment_variables import SECONDS_IN_MINUTE

from collections import defaultdict, deque


class simple_utc(tzinfo):
    def tzname(self, **kwargs):
        return "UTC"

    def utcoffset(self, dt):
        return timedelta(0)


def convert_dict(key_type, value_type, dict_item) -> dict:
    if dict_item is None:
        return dict()
    return {
        key_type(k): value_type(v)
        for k, v in sorted(dict_item.items(), key=lambda i: key_type(i[0]))
    }


def find_effective_value_for_time_step(time_step: int, schedule: dict) -> float:
    if time_step in schedule:
        return schedule[time_step]

    effective_time_step = max(int(k) for k in schedule.keys() if int(k) <= time_step)
    return schedule.get(str(effective_time_step), 0.0)


class FromDictToSimulatorConverter(Timer):
    def __init__(self, reader: FolderReader, writer: FolderWriter):
        Timer.__init__(self, prefix="[Convert]")

        self.reader = reader
        self.writer = writer

        self.configuration = self.reader.files.get("Configuration", [])[0]

        self.total_time_steps = (
            self.configuration["StepsPerCycle"] * self.configuration["SimulatedCycles"]
        )

        self.files_content = dict()

        self.current_group = 0

        for key in simulator_format.keys():
            self.files_content[key] = list()

        self.entity_order = dict()

        self.lead_times = {}

    def __scheduler(self):
        inputs = self.reader.files.get("input", [])
        outputs = self.reader.files.get("output", [])
        contains = self.reader.files.get("contains", [])
        transports = self.reader.files.get("Transport", [])
        stocks = self.reader.files.get("Stock", [])

        operations_names = set(i["target"] for i in inputs)
        transports_names = set(
            t["Label"] for t in transports
        )  # TODO to be replaced by RelationshipId when available from azure-digital-twins-simulator-connector

        stocks_names = set(stock["id"] for stock in stocks)
        _input_stocks = set(i["source"] for i in inputs) | set(
            t["source"] for t in transports
        )
        _output_stocks = set(o["target"] for o in outputs) | set(
            t["target"] for t in transports
        )

        start_stocks = _input_stocks - _output_stocks
        final_stocks = _output_stocks - _input_stocks

        scheduled_elements = []
        for stock in sorted(start_stocks):
            scheduled_element = dict(Type="Rule", Name=stock, TaskName="Dispatch")
            scheduled_elements.append(scheduled_element)
        self.files_content["IndustrialNetworkScheduler"].extend(scheduled_elements)

        stocks_queue = deque()

        contractors = set(
            op["id"]
            for op in self.reader.files.get("ProductionOperation", [])
            if op["IsContractor"]
        )

        for st in final_stocks:
            stocks_queue.append(st)
            self.entity_order[st] = 0

        parents = defaultdict(set)

        # While we have stocks to treat we continue
        for links in [inputs, outputs]:
            for link in links:
                parents[link["source"]].add(link["target"])
        for tr in transports:
            parents[tr["source"]].add(tr["Label"])
            parents[tr["Label"]].add(tr["target"])

        for production_resource in [c["source"] for c in contains]:
            brotherhood = [
                c["target"] for c in contains if c["source"] == production_resource
            ]
            first_child = brotherhood.pop()
            for child in brotherhood:
                parents[first_child] |= parents[child]
            for child in brotherhood:
                parents[child] = parents[first_child]

        max_depth = 0

        def get_entity_order(_entity):
            if _entity not in self.entity_order:
                self.entity_order[_entity] = 1 + max(
                    map(get_entity_order, parents[_entity])
                )
            return self.entity_order[_entity]

        for entity in parents.keys():
            o = get_entity_order(entity)
            max_depth = max(max_depth, o)

        operation_order = [[] for _ in range(max_depth)]
        for k, v in self.entity_order.items():
            if k in transports_names or k in operations_names:
                operation_order[v].append(k)

        operation_order.reverse()
        operation_order = [i for i in operation_order if len(i) > 0]

        scheduled_ComputeStockToTransfer_stocks = set()

        for list_operations in operation_order:
            # For each rank of operation, every operation can be done in parallel, so order is not important

            # To schedule we need some entities and information about them:

            fixed_ee = set()
            # StockToOperation edge entities linked to the contractor Operation
            # entites considered at this rank.

            prop_ee = set()
            # StockToOperation edge entities linked to the regular Operation
            # entites considered at this rank.

            all_fixed_ee = set()
            # StockToOperation edge entities linked to the Stock entities which
            # are taken as input by the contractor Operation entities considered
            #  at this rank and for which ComputeStockToTransfer processes were
            # not scheduled yet.

            all_prop_ee = set()
            # StockToOperation edge entities linked to the Stock entities which
            #  are taken as input by the regular Operation entities considered at
            # this rank and for which ComputeStockToTransfer processes were not
            # scheduled yet.

            input_stocks = set()
            # Stock entities taken as input by the Operation entities considered
            # at this rank and for which ComputeStockToTransfer processes were
            # not scheduled yet.

            machines = set()
            # Machine entities containing the non transport Operation entities
            #  considered at this rank.

            operation_ee = set()
            # StockToOperation edge entities for the non transport Operation
            # entities considered at this rank.

            transport_ops = set()
            # Operation entities considered at this rank which are transports.

            end_stocks = set()
            # Stock entities outputed by the Operation entities considered at
            # this rank.

            for operation in list_operations:
                if operation in operations_names:
                    edge_entities = set(
                        i["source"] + "_to_" + i["target"]
                        for i in inputs
                        if i["target"] == operation
                    )
                    machines.add(
                        next(c["source"] for c in contains if c["target"] == operation)
                    )
                    stock = next(
                        o["target"] for o in outputs if o["source"] == operation
                    )
                    operation_ee.update(edge_entities)
                    input_stocks |= set(
                        (i["source"] for i in inputs if i["target"] == operation)
                    )
                else:
                    edge_entities = set(
                        t["source"] + "_to_" + t["Label"]
                        for t in transports
                        if t["Label"] == operation
                    )
                    transport_ops.add(operation)
                    stock = next(
                        t["target"] for t in transports if t["Label"] == operation
                    )
                    input_stocks |= set(
                        (t["source"] for t in transports if t["Label"] == operation)
                    )

                if operation in contractors:
                    fixed_ee.update(edge_entities)
                else:
                    prop_ee.update(edge_entities)

                end_stocks.add(stock)

            input_stocks -= scheduled_ComputeStockToTransfer_stocks
            if input_stocks:
                scheduled_ComputeStockToTransfer_stocks |= input_stocks
                for i in inputs:
                    if i["source"] in input_stocks:
                        operation = i["target"]
                        (all_fixed_ee if operation in contractors else all_prop_ee).add(
                            i["source"] + "_to_" + operation
                        )
                for t in transports:
                    if t["source"] in input_stocks:
                        operation = t["Label"]
                        (all_fixed_ee if operation in contractors else all_prop_ee).add(
                            t["source"] + "_to_" + operation
                        )

            # Names must be sorted to stabilize the output schedule.
            all_fixed_ee = sorted(all_fixed_ee)
            fixed_ee = sorted(fixed_ee)
            all_prop_ee = sorted(all_prop_ee)
            prop_ee = sorted(prop_ee)
            machines = sorted(machines)
            operation_ee = sorted(operation_ee)
            transport_ops = sorted(transport_ops)
            end_stocks = sorted(end_stocks)

            # Now we can create our scheduled Processes / Rules
            scheduled_elements = []

            # As soon as a ComputeStockToTransfer process is scheduled for a
            # particular stock, all ComputeStockToTransfer processes for the
            # said stock should be scheduled at the same time. Moreover, they
            # should be scheduled only once.
            for EE in all_fixed_ee:
                scheduled_element = dict(
                    Type="Process", Name=EE, TaskName="ComputeStockToTransfer"
                )
                scheduled_elements.append(scheduled_element)

            for EE in fixed_ee:
                scheduled_element = dict(
                    Type="Process", Name=EE, TaskName="TransferStock"
                )
                scheduled_elements.append(scheduled_element)

            for EE in all_prop_ee:
                scheduled_element = dict(
                    Type="Process", Name=EE, TaskName="ComputeStockToTransfer"
                )
                scheduled_elements.append(scheduled_element)

            for EE in prop_ee:
                scheduled_element = dict(
                    Type="Process", Name=EE, TaskName="TransferStock"
                )
                scheduled_elements.append(scheduled_element)

            for mac in machines:
                scheduled_element = dict(Type="Rule", Name=mac, TaskName="SelectStock")
                scheduled_elements.append(scheduled_element)

            for EE in operation_ee:
                scheduled_element = dict(
                    Type="Process", Name=EE, TaskName="SendBackStockIfNecessary"
                )
                scheduled_elements.append(scheduled_element)

            for mac in machines:
                scheduled_element = dict(
                    Type="Process", Name=mac, TaskName="LaunchProcessing"
                )
                scheduled_elements.append(scheduled_element)

            for transport in transport_ops:
                scheduled_element = dict(
                    Type="Rule", Name=transport, TaskName="SelectStock"
                )
                scheduled_elements.append(scheduled_element)

            for transport in transport_ops:
                scheduled_element = dict(
                    Type="Process", Name=transport, TaskName="Process"
                )
                scheduled_elements.append(scheduled_element)

            for stock in end_stocks:
                scheduled_element = dict(
                    Type="Rule", Name=stock, TaskName="ReceiveStock"
                )
                scheduled_elements.append(scheduled_element)

            for stock in end_stocks:
                scheduled_element = dict(Type="Rule", Name=stock, TaskName="Dispatch")
                scheduled_elements.append(scheduled_element)

            self.files_content["IndustrialNetworkScheduler"].extend(scheduled_elements)
        self.files_content["IndustrialNetworkScheduler"].append(
            dict(
                Type="Process", Name="IndustrialNetwork", TaskName="MesoResolveDemands"
            )
        )
        self.files_content["IndustrialNetworkScheduler"].extend(
            (
                dict(Type="Rule", Name=stock, TaskName="CheckLevelAndOrderIfNecessary")
                for stock in sorted(
                    sorted(stocks_names), key=lambda e: self.entity_order.get(e, -1)
                )
            )
        )
        self.writer.write_from_list(
            self.files_content["IndustrialNetworkScheduler"],
            "IndustrialNetworkScheduler",
            ordering_key=None,
        )

    def __lead_times(self):
        inputs = self.reader.files.get("input", [])
        operations = self.reader.files.get("ProductionOperation", [])
        outputs = self.reader.files.get("output", [])
        stocks = self.reader.files.get("Stock", [])
        transports = self.reader.files.get("Transport", [])

        stock_names = set(stock["id"] for stock in stocks)

        non_input_stock_names = (
            stock_names
            - set(i["source"] for i in inputs)
            - set(transport["source"] for transport in transports)
        )
        non_output_stock_names = (
            stock_names
            - set(output["target"] for output in outputs)
            - set(transport["target"] for transport in transports)
        )

        parents = defaultdict(set)
        for links in [inputs, outputs]:
            for link in links:
                parents[link["target"]].add(link["source"])
        for transport in transports:
            parents[transport["Label"]].add(transport["source"])
            parents[transport["target"]].add(transport["Label"])

        durations = (
            {stock["id"]: 0 for stock in stocks}
            | {operation["id"]: operation["Duration"] for operation in operations}
            | {transport["Label"]: transport["Duration"] for transport in transports}
        )

        for stock_name in non_output_stock_names:
            self.lead_times[stock_name] = 0

        def get_lead_time(entity):
            if entity not in self.lead_times:
                self.lead_times[entity] = (
                    max(get_lead_time(parent) for parent in parents[entity])
                    + durations[entity]
                )
            return self.lead_times[entity]

        for stock_name in non_input_stock_names:
            get_lead_time(stock_name)

    def __operation(self):

        operations_csv = self.reader.files.get("ProductionOperation", [])
        contains_csv = self.reader.files.get("contains", [])
        outputs = self.reader.files.get("output", [])
        stock_csv = self.reader.files.get("Stock", [])

        for operation in operations_csv:
            parent_resource = next(
                c["source"] for c in contains_csv if c["target"] == operation["id"]
            )
            output_stock = next(
                link["target"] for link in outputs if link["source"] == operation["id"]
            )
            follows_stock_policy = next(
                stock["StockPolicy"] != "None"
                for stock in stock_csv
                if stock["id"] == output_stock
            )
            operation_item = {
                "Name": operation["id"],
                "ParentName": parent_resource,
                "OperatingPerformanceSchedule": convert_dict(
                    int, float, operation["OperatingPerformances"]
                ),
                "CycleTimeSchedule": convert_dict(int, float, operation["CycleTimes"]),
                "RejectRateSchedule": convert_dict(
                    int, float, operation["RejectRates"]
                ),
                "OperatingPerformanceUncertaintySchedule": convert_dict(
                    int, float, operation["OperatingPerformanceUncertainties"]
                ),
                "ProductionUnitCostSchedule": convert_dict(
                    int, float, operation["ProductionUnitCosts"]
                ),
                "CO2UnitEmissionsSchedule": convert_dict(
                    int, float, operation["CO2UnitEmissions"]
                ),
                "ProductionPlanSchedule": convert_dict(
                    int, float, operation["QuantitiesToProduce"]
                ),
                "MinimumOrderQuantitySchedule": convert_dict(
                    int, float, operation["MinimumOrderQuantities"]
                ),
                "MultipleOrderQuantitySchedule": convert_dict(
                    int, float, operation["MultipleOrderQuantities"]
                ),
                "SourcingProportionSchedule": convert_dict(
                    int, float, operation["SourcingProportions"]
                ),
                "InvestmentCost": operation["InvestmentCost"],
                "IsContractor": operation["IsContractor"],
                "Priority": operation["Priority"],
                "SchedulingStep": self.entity_order.get(operation["id"], -1),
                "FollowsStockPolicy": follows_stock_policy,
                "Duration": operation["Duration"],
                "LeadTime": self.lead_times[operation["id"]],
            }
            self.files_content["BE_Operation"].append(operation_item)
        self.writer.write_from_list(
            self.files_content["BE_Operation"], "BE_Operation", ordering_key="Name"
        )

    def __model_parameters(self):

        production_plan_is_available = all(
            bool(po["QuantitiesToProduce"])
            for po in self.reader.files.get("ProductionOperation", [])
        )
        model_parameters = {
            "TimeStepDuration": seconds_to_iso_8061_duration(
                self.configuration["TimeStepDuration"] * SECONDS_IN_MINUTE
            ),
            "TimeStepPerCycle": self.configuration["StepsPerCycle"],
            "NumberOfCycle": self.configuration["SimulatedCycles"],
            "StartingDate": datetime.fromisoformat(self.configuration["StartingDate"])
            .replace(tzinfo=simple_utc())
            .isoformat(),
            "ManageBacklogQuantities": self.configuration["ManageBacklogQuantities"],
            "OptimizationObjective": self.configuration["OptimizationObjective"],
            "ActivateUncertainties": self.configuration["ActivateUncertainties"],
            "EmptyObsoleteStocks": self.configuration["EmptyObsoleteStocks"],
            "UncertaintiesProbabilityDistribution": self.configuration[
                "UncertaintiesProbabilityDistribution"
            ],
            "TransportUncertaintiesProbabilityDistribution": self.configuration[
                "TransportUncertaintiesProbabilityDistribution"
            ],
            "ActivateVariableMachineOpeningRate": self.configuration[
                "ActivateVariableMachineOpeningRate"
            ],
            "ImmobilizedCashRelativeCost": self.configuration["FinancialCostOfStock"],
            "BatchSize": self.configuration["BatchSize"],
            "ExtraProductionPlanIsAvailable": production_plan_is_available,
            "EnforceProductionPlan": self.configuration["EnforceProductionPlan"],
            "AllowPartRetention": self.configuration["IntermediaryStockDispatchPolicy"]
            == "AllowRetention",
            "ActualizeShipments": self.configuration["ActualizeShipments"],
            "ActivateCorrelatedDemandUncertainties": self.configuration[
                "ActivateCorrelatedDemandUncertainties"
            ],
            "DemandCorrelations": self.configuration["DemandCorrelations"],
        }

        self.files_content["ModelParameters"].append(model_parameters)
        self.writer.write_from_list(
            self.files_content["ModelParameters"], "ModelParameters", ordering_key=None
        )

    def __stock(self):
        stock_csv = self.reader.files.get("Stock", [])
        for stock in stock_csv:
            maximal_stock = stock["MaximalStock"]
            stock_item = {
                "Name": stock["id"],
                "CurrentStock": stock["InitialStock"],
                "MinimalStock": stock["MinimalStock"],
                "MaximalStock": maximal_stock if maximal_stock is not None else -1,
                "Value": stock["InitialValue"],
                "PurchasingUnitCostSchedule": convert_dict(
                    int, float, stock["PurchasingUnitCosts"]
                ),
                "CO2UnitEmissionsSchedule": convert_dict(
                    int, float, stock["CO2UnitEmissions"]
                ),
                "UnitIncomeSchedule": convert_dict(int, float, stock["UnitIncomes"]),
                "Demand": dict(),
                "StorageUnitCostSchedule": convert_dict(
                    int, float, stock["StorageUnitCosts"]
                ),
                "IsInfinite": stock["IsInfinite"],
                "StockPolicy": stock["StockPolicy"],
                "SourcingPolicy": stock["SourcingPolicy"],
                "DispatchPolicy": stock["DispatchPolicy"],
                "ReviewPeriod": stock["ReviewPeriod"],
                "FirstReview": stock["FirstReview"],
                "Advance": stock["Advance"],
                "LeadTime": self.lead_times[stock["id"]],
                "SalesForecasts": convert_dict(int, float, stock["SalesForecasts"]),
                "OrderPointSchedule": convert_dict(int, float, stock["OrderPoints"]),
                "OrderQuantitySchedule": dict(),
                "SafetyQuantitySchedule": convert_dict(
                    int, float, stock["SafetyQuantities"]
                ),
                "SchedulingStep": self.entity_order.get(stock["id"], -1),
            }

            if stock_item["StockPolicy"] == "OrderPointFixedQuantity":
                stock_item["OrderQuantitySchedule"] = convert_dict(
                    int, float, stock["OrderQuantities"]
                )
            elif stock_item["StockPolicy"] == "OrderPointVariableQuantity":
                stock_item["OrderQuantitySchedule"] = convert_dict(
                    int, float, stock["OrderUpToLevels"]
                )

            backlog_weight = stock["BacklogWeight"]
            last_uncertainty = 0

            if stock["Demands"]:
                demands = convert_dict(int, float, stock["Demands"])
                demands_weight = convert_dict(int, float, stock["DemandWeights"])
                demands_uncertainties = convert_dict(
                    int, float, stock["DemandUncertainties"]
                )

                last_weight = 0

                for time_step in range(self.total_time_steps):
                    if time_step in demands_weight:
                        last_weight = demands_weight[time_step]
                    if time_step in demands_uncertainties:
                        last_uncertainty = demands_uncertainties[time_step]
                    if time_step in demands and demands[time_step]:
                        current_demand = dict(
                            ExternalDemand=demands[time_step],
                            InternalDemand=0.0,
                            BacklogWeight=backlog_weight,
                            ExternalWeight=last_weight,
                            InternalWeight=0.0,
                            WeightMax=0.0,
                            MaxVal=0.0,
                            DemandRelativeUncertainty=last_uncertainty,
                        )
                        stock_item["Demand"][time_step] = current_demand

            maximization_weight = stock["MaximizationWeight"]
            if maximization_weight:
                total_max_production = 0
                for time_step in range(self.total_time_steps):
                    total_max_production += self.compute_theoretical_maximum_production(
                        stock["id"], time_step
                    )
                if self.total_time_steps - 1 not in stock_item["Demand"]:
                    current_demand = dict(
                        ExternalDemand=0.0,
                        InternalDemand=0.0,
                        BacklogWeight=backlog_weight,
                        ExternalWeight=0.0,
                        InternalWeight=0.0,
                        WeightMax=0.0,
                        MaxVal=0.0,
                        DemandRelativeUncertainty=last_uncertainty,
                    )
                    stock_item["Demand"][self.total_time_steps - 1] = current_demand
                stock_item["Demand"][self.total_time_steps - 1][
                    "WeightMax"
                ] = maximization_weight
                stock_item["Demand"][self.total_time_steps - 1][
                    "MaxVal"
                ] = total_max_production

            self.files_content["BE_Stock"].append(stock_item)
        self.writer.write_from_list(
            self.files_content["BE_Stock"], "BE_Stock", ordering_key="Name"
        )

    def compute_theoretical_maximum_production(
        self, stock_id: str, time_step: int
    ) -> float:
        stock = next(
            s for s in self.reader.files.get("Stock", []) if s["id"] == stock_id
        )
        theoretical_maximum = stock["InitialStock"] if time_step == 0 else 0.0
        transports = [
            transport
            for transport in self.reader.files.get("Transport", [])
            if transport["target"] == stock_id
        ]
        for transport in transports:
            origin = transport["source"]
            duration = transport["Duration"]
            theoretical_maximum += transport["InitialTransportedQuantities"].get(
                str(time_step), 0.0
            )
            if time_step >= duration:
                theoretical_maximum += self.compute_theoretical_maximum_production(
                    origin, time_step - duration
                )
        operation_names = set(
            output["source"]
            for output in self.reader.files.get("output", [])
            if output["target"] == stock_id
        )
        operations = [
            operation
            for operation in self.reader.files.get("ProductionOperation", [])
            if operation["id"] in operation_names
        ]
        for operation in operations:
            if operation["IsContractor"]:
                theoretical_maximum += find_effective_value_for_time_step(
                    time_step, operation["QuantitiesToProduce"]
                )
            else:
                pr_name = next(
                    contains["source"]
                    for contains in self.reader.files.get("contains", [])
                    if contains["target"] == operation["id"]
                )
                pr_data = next(
                    pr
                    for pr in self.reader.files.get("ProductionResource", [])
                    if pr["id"] == pr_name
                )
                max_opening_time = find_effective_value_for_time_step(
                    time_step, pr_data["OpeningTimes"]
                )
                ct = (
                    find_effective_value_for_time_step(
                        time_step, operation["CycleTimes"]
                    )
                    if operation["CycleTimes"]
                    else variables_default_values["ProductionOperation"]["CycleTimes"]
                )
                rr = (
                    find_effective_value_for_time_step(
                        time_step, operation["RejectRates"]
                    )
                    if operation["RejectRates"]
                    else variables_default_values["ProductionOperation"]["RejectRates"]
                )
                op = (
                    find_effective_value_for_time_step(
                        time_step, operation["OperatingPerformances"]
                    )
                    if operation["OperatingPerformances"]
                    else variables_default_values["ProductionOperation"][
                        "OperatingPerformances"
                    ]
                )
                theoretical_maximum += (max_opening_time / ct) * op * (1.0 - rr)
        return theoretical_maximum

    def __machine(self):
        pr_csv = self.reader.files.get("ProductionResource", [])
        contains_csv = self.reader.files.get("contains", [])

        for pr in pr_csv:
            target_operations = [
                c["target"] for c in contains_csv if c["source"] == pr["id"]
            ]
            operations_count = len(target_operations)
            target_propotions = {
                op: 1.0 / operations_count for op in sorted(target_operations)
            }

            group = {}
            if operations_count > 1:
                group = {
                    self.current_group: {
                        "StartCycle": 0,
                        "EndCycle": self.configuration["SimulatedCycles"] - 1,
                    }
                }
                self.current_group += 1

            pr_item = {
                "Name": pr["id"],
                "StockTypeProportion": [target_propotions]
                * self.configuration["SimulatedCycles"],
                "Groups": group,
                "OpeningRates": [1] * self.configuration["SimulatedCycles"],
                "OpeningTimeSchedule": convert_dict(int, float, pr["OpeningTimes"]),
                "ProductionCostSchedule": convert_dict(
                    int, float, pr["FixedProductionCosts"]
                ),
                "ProductionPolicy": pr["ProductionPolicy"],
            }
            self.files_content["CE_Machine"].append(pr_item)
        self.writer.write_from_list(
            self.files_content["CE_Machine"], "CE_Machine", ordering_key="Name"
        )

    def __op_to_stock(self):
        output_csv = self.reader.files.get("output", [])
        for o in output_csv:
            op_to_stock_item = {"StockName": o["target"], "OperationName": o["source"]}
            self.files_content["Arcs_OperationToStock"].append(op_to_stock_item)
        output_csv = self.reader.files.get("Transport", [])
        for o in output_csv:
            op_to_stock_item = {"StockName": o["target"], "OperationName": o["Label"]}
            self.files_content["Arcs_OperationToStock"].append(op_to_stock_item)
        self.writer.write_from_list(
            self.files_content["Arcs_OperationToStock"],
            "Arcs_OperationToStock",
            ordering_key="StockName",
        )

    def __transport(self):
        transport_csv = self.reader.files.get("Transport", [])
        stock_csv = self.reader.files.get("Stock", [])

        def dict_to_list(duration: int, data: dict):
            if data:
                return [data.get(str(t), 0.0) for t in range(duration)]
            return [0.0 for _ in range(duration)]

        for transport in transport_csv:
            follows_stock_policy = next(
                stock["StockPolicy"] != "None"
                for stock in stock_csv
                if stock["id"] == transport["target"]
            )
            transport_item = {
                "Name": transport["Label"],
                # TODO to be replaced by RelationshipId when available from azure-digital-twins-simulator-connector
                "Duration": transport["Duration"],
                "Shipments": {},
                "TransportUnitCostSchedule": convert_dict(
                    int, float, transport["TransportUnitCosts"]
                ),
                "CustomsUnitCostSchedule": convert_dict(
                    int, float, transport["CustomFees"]
                ),
                "CO2UnitEmissionsSchedule": convert_dict(
                    int, float, transport["CO2UnitEmissions"]
                ),
                "ActualDurationSchedule": convert_dict(
                    int, int, transport["ActualDurations"]
                ),
                "MinimumOrderQuantitySchedule": convert_dict(
                    int, float, transport["MinimumOrderQuantities"]
                ),
                "MultipleOrderQuantitySchedule": convert_dict(
                    int, float, transport["MultipleOrderQuantities"]
                ),
                "SourcingProportionSchedule": convert_dict(
                    int, float, transport["SourcingProportions"]
                ),
                "TransportUncertaintiesParameter1": convert_dict(
                    int, float, transport["TransportUncertaintiesParameter1"]
                ),
                "TransportUncertaintiesParameter2": convert_dict(
                    int, float, transport["TransportUncertaintiesParameter2"]
                ),
                "TransportUncertaintiesParameter3": convert_dict(
                    int, float, transport["TransportUncertaintiesParameter3"]
                ),
                "TransportUncertaintiesParameter4": convert_dict(
                    int, float, transport["TransportUncertaintiesParameter4"]
                ),
                "Priority": transport["Priority"],
                "SchedulingStep": self.entity_order.get(transport["Label"], -1),
                "FollowsStockPolicy": follows_stock_policy,
                "LeadTime": self.lead_times[transport["Label"]],
            }
            initial_quantities = convert_dict(
                int, float, transport["InitialTransportedQuantities"]
            )
            initial_values = convert_dict(
                int, float, transport["InitialTransportedValues"]
            )
            co2_unit_emissions = transport_item["CO2UnitEmissionsSchedule"].get(
                0, variables_default_values["Transport"]["CO2UnitEmissions"]
            )
            default_value = variables_default_values["Transport"][
                "InitialTransportedValues"
            ]
            time_steps = reversed(
                [
                    (-i, t)
                    for i, t in enumerate(
                        sorted(
                            (t for t, q in initial_quantities.items() if 0.0 < q),
                            reverse=True,
                        ),
                        1,
                    )
                ]
            )
            for departure, arrival in time_steps:
                quantity = initial_quantities[arrival]
                transport_item["Shipments"][departure] = {
                    "Arrival": arrival,
                    "ActualArrival": arrival,
                    "Quantity": quantity,
                    "Value": initial_values.get(arrival, default_value),
                    "CO2Emissions": quantity * co2_unit_emissions,
                }
            self.files_content["BE_TransportOperation"].append(transport_item)

        self.writer.write_from_list(
            self.files_content["BE_TransportOperation"],
            "BE_TransportOperation",
            ordering_key="Name",
        )

    def __stock_to_operation(self):
        input_csv = self.reader.files.get("input", [])
        transport_csv = self.reader.files.get("Transport", [])

        number_of_edges_by_stock = defaultdict(int)

        for i in input_csv:
            number_of_edges_by_stock[i["source"]] += 1

        for t in transport_csv:
            number_of_edges_by_stock[t["source"]] += 1

        groups = dict()

        stock_has_demand = {
            s["id"]: bool(s["Demands"])  # DS
            for s in self.reader.files.get("Stock", [])
        }

        for stock in sorted(number_of_edges_by_stock.keys()):
            groups[stock] = {}
            if number_of_edges_by_stock[stock] > 1 or stock_has_demand[stock]:
                groups[stock] = {
                    self.current_group: {
                        "StartCycle": 0,
                        "EndCycle": self.configuration["SimulatedCycles"] - 1,
                    }
                }
                self.current_group += 1

        contractors = set(
            op["id"]
            for op in self.reader.files.get("ProductionOperation", [])
            if op["IsContractor"]
        )

        for i in input_csv:
            prop_sto = [1 / number_of_edges_by_stock[i["source"]]] * self.configuration[
                "SimulatedCycles"
            ]
            sto_item = {
                "Name": i["source"] + "_to_" + i["target"],
                "OperationName": i["target"],
                "StockName": i["source"],
                "PropStockForOperation": prop_sto,
                "IsProportion": i["target"] not in contractors,
                "Groups": groups[i["source"]],
                "Cardinality": i["InputQuantity"],
            }
            self.files_content["EE_StockToOperation"].append(sto_item)

        for i in transport_csv:
            prop_sto = [1 / number_of_edges_by_stock[i["source"]]] * self.configuration[
                "SimulatedCycles"
            ]
            sto_item = {
                "Name": i["source"] + "_to_" + i["Label"],
                "OperationName": i["Label"],
                # TODO to be replaced by RelationshipId when available from azure-digital-twins-simulator-connector
                "StockName": i["source"],
                "PropStockForOperation": prop_sto,
                "IsProportion": True,
                "Groups": groups[i["source"]],
                "Cardinality": 1,
            }
            self.files_content["EE_StockToOperation"].append(sto_item)

        self.writer.write_from_list(
            self.files_content["EE_StockToOperation"],
            "EE_StockToOperation",
            ordering_key="Name",
        )

    def transform(self):
        self.__scheduler()
        self.split(
            "{time_since_last_split:6.4f} ({time_since_start:6.4f}) - IndustrialNetworkScheduler"
        )
        self.__lead_times()
        self.split(
            "{time_since_last_split:6.4f} ({time_since_start:6.4f}) - Lead times"
        )
        self.__stock()
        self.split("{time_since_last_split:6.4f} ({time_since_start:6.4f}) - BE_Stock")
        self.__operation()
        self.split(
            "{time_since_last_split:6.4f} ({time_since_start:6.4f}) - BE_Operation"
        )
        self.__transport()
        self.split(
            "{time_since_last_split:6.4f} ({time_since_start:6.4f}) - BE_TransportOperation"
        )
        self.__machine()
        self.split(
            "{time_since_last_split:6.4f} ({time_since_start:6.4f}) - CE_Machine"
        )
        self.__op_to_stock()
        self.split(
            "{time_since_last_split:6.4f} ({time_since_start:6.4f}) - Arcs_OperationToStock"
        )
        self.__stock_to_operation()
        self.split(
            "{time_since_last_split:6.4f} ({time_since_start:6.4f}) - EE_StockToOperation"
        )
        self.__model_parameters()
        self.split(
            "{time_since_last_split:6.4f} ({time_since_start:6.4f}) - ModelParameters"
        )



from CosmoTech_Acceleration_Library.Accelerators.adx_wrapper import ADXQueriesWrapper
from Supplychain.Generic.adx_wrapper import ADXWrapper
from Supplychain.Generic.folder_io import FolderWriter

from typing import Union


class ADXAndFileWriter():

    def write_target_file(self, dict_list: list, file_name: str, drop_by_tag: str = None):
        """
        Will wirte the data in the required format :
        - on disk if required, as csv files
        - on ADX if required
        :param dict_list: a list of dict objects to write
        :param file_name: the name of the file (without extension) which will be used as table name on ADX
        :return: None
        """
        if self.writer is not None:
            self.writer.write_from_list(dict_list=dict_list,
                                        file_name=file_name,
                                        ordering_key=None)
        if self.adx_connector:
            self.adx_connector.send_to_adx(dict_list, file_name, drop_by_tag=drop_by_tag)

    def __init__(self,
                 writer: Union[FolderWriter, None] = None,
                 adx_connector: Union[ADXQueriesWrapper, ADXWrapper, None] = None):
        """
        Init and execute transformation of the data
        :param reader: Folder reader serving files
        :param writer: Potential folder writer
        :param adx_connector: Potential connector to ADX
        """
        self.adx_connector = adx_connector

        self.writer = writer



import csv
import glob
import io
import json
import os

from collections import defaultdict

from Supplychain.Generic.folder_io import FolderReader


class CSVReader(FolderReader):
    extension = ".csv"

    def refresh(self):
        filenames = glob.glob(os.path.join(self.input_folder, "*" + self.extension))
        self.files = defaultdict(list)

        for filename in filenames:
            with open(filename, "r") as file:
                # Read every file in the input folder
                current_filename = os.path.basename(filename)[:-len(self.extension)]
                for row in csv.DictReader(file):
                    new_row = dict()
                    for key, value in row.items():
                        if key is None:
                            continue
                        try:
                            # Try to convert any json row to dict object
                            converted_value = json.load(io.StringIO(value))
                        except json.decoder.JSONDecodeError:
                            converted_value = value
                        if converted_value == '':
                            converted_value = None
                        if converted_value is not None or self.keep_nones:
                            new_row[key] = converted_value
                    self.files[current_filename].append(new_row)

    def __init__(self,
                 input_folder: str = "Input",
                 keep_nones: bool = True):

        FolderReader.__init__(self,
                              input_folder=input_folder,
                              keep_nones=keep_nones)

        self.refresh()
